import concurrent.futures
import time
from unittest import TestCase

from flask import Flask

from dm.web.extensions.flask_executor.executor import Executor
from dm.web.extensions.flask_executor.futures import FutureCollection, FutureProxy
from dm.web.extensions.flask_executor.helpers import InstanceProxy


def fib(n):
    if n <= 2:
        return 1
    else:
        return fib(n - 1) + fib(n - 2)


class TestFutures(TestCase):

    def test_plain_future(self):
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        futures = FutureCollection()
        future = executor.submit(fib, 33)
        futures.add('fibonacci', future)
        assert futures.done('fibonacci') is False
        assert futures._state('fibonacci') is not None
        assert future in futures
        futures.pop('fibonacci')
        assert future not in futures

    def test_missing_future(self):
        futures = FutureCollection()
        assert futures.running('test') is None

    def test_duplicate_add_future(self):
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        futures = FutureCollection()
        future = executor.submit(fib, 33)
        futures.add('fibonacci', future)
        try:
            futures.add('fibonacci', future)
        except ValueError:
            assert True
        else:
            assert False

    def test_futures_max_length(self):
        executor = concurrent.futures.ThreadPoolExecutor(max_workers=1)
        futures = FutureCollection(max_length=10)
        future = executor.submit(pow, 2, 4)
        futures.add(0, future)
        assert future in futures
        assert len(futures) == 1
        for i in range(1, 11):
            futures.add(i, executor.submit(pow, 2, 4))
        assert len(futures) == 10
        assert future not in futures

    def test_future_proxy(self):
        app = Flask(__name__)
        executor = Executor(app)
        with app.test_request_context(''):
            future = executor.submit(pow, 2, 4)
        # Test if we're returning a subclass of Future
        assert isinstance(future, concurrent.futures.Future)
        assert isinstance(future, FutureProxy)
        concurrent.futures.wait([future])
        # test standard Future methods and attributes
        assert future._state == concurrent.futures._base.FINISHED
        assert future.done()
        assert future.exception(timeout=0) is None

    def test_add_done_callback(self):
        """Exceptions thrown in callbacks can't be easily caught and make it hard
        to test for callback failure. To combat this, a global variable is used to
        store the value of an exception and test for its existence.
        """
        app = Flask(__name__)
        executor = Executor(app)
        global exception
        exception = None
        with app.test_request_context(''):
            future = executor.submit(time.sleep, 0.5)

            def callback(future):
                global exception
                try:
                    executor.submit(time.sleep, 0)
                except RuntimeError as e:
                    exception = e

            future.add_done_callback(callback)
        concurrent.futures.wait([future])
        assert exception is None

    def test_instance_proxy(self):
        class TestProxy(InstanceProxy):
            pass

        x = TestProxy(concurrent.futures.Future())
        assert isinstance(x, concurrent.futures.Future)
        assert 'TestProxy' in repr(x)
        assert 'Future' in repr(x)



import os
import sys

from dm.dshell import environ
from dm.dshell.history import FileTagHistory
from dm.dshell.network import get, get_parameters_from_path


def exit_dshell(msg='Goodbye!', rc=0):
    print(msg)
    sys.exit(rc)


def get_history(tag, default=None):
    file_history = environ.get('FILE_HISTORY', None)
    if file_history:
        file_history = os.path.expanduser('~/.dshell_history')
        if not os.path.exists(file_history):
            open(file_history, 'w').close()
        history = FileTagHistory(file_history, tag)
    else:
        history = default
    return history


def id2name(view, iden, key='name'):
    view_args = {p: iden for p in get_parameters_from_path(view)}
    if len(view_args) != 1:
        raise ValueError('view has more than 1 parameter to resolve')
    resp = get(view, view_args)
    if resp.code and 200 <= resp.code <= 299:
        if callable(key):
            return key(resp.msg)
        else:
            return resp.msg.get('key')


def name2id(view, name, key='name'):
    view_args = {f'filter[{key}]': name}
    resp = get(view, view_args)
    if resp.ok:
        if len(resp.msg) > 1:
            raise ValueError('multiple ids found')
        return resp.msg[0].get('id', None)



import datetime
import re
import typing as t
import uuid
from collections import ChainMap

from jinja2schema import infer
from sqlalchemy import orm

from dm import defaults
from dm.domain.entities.base import UUIDistributedEntityMixin
from dm.utils.typos import UUID, ScalarListType, UtcDateTime
from dm.web import db, errors
from .action_template import ActionType
from ...utils import typos
from ...utils.helpers import get_now

if t.TYPE_CHECKING:
    from .action_template import ActionTemplate
    from .orchestration import Orchestration

step_step = db.Table('D_step_step',
                     db.Column('parent_step_id', UUID, db.ForeignKey('D_step.id'), primary_key=True),
                     db.Column('step_id', UUID, db.ForeignKey('D_step.id'), primary_key=True),
                     )


class Step(db.Model, UUIDistributedEntityMixin):
    __tablename__ = "D_step"
    order = 30
    id = db.Column(UUID, primary_key=True, default=uuid.uuid4)
    orchestration_id = db.Column(UUID, db.ForeignKey('D_orchestration.id'), nullable=False)
    action_template_id = db.Column(UUID, db.ForeignKey('D_action_template.id'))
    undo = db.Column(db.Boolean, nullable=False)
    step_stop_on_error = db.Column("stop_on_error", db.Boolean)
    step_stop_undo_on_error = db.Column("stop_undo_on_error", db.Boolean)
    step_undo_on_error = db.Column("undo_on_error", db.Boolean)
    step_parameters = db.Column("parameters", db.JSON)
    step_expected_stdout = db.Column("expected_stdout", db.Text)
    step_expected_stderr = db.Column("expected_stderr", db.Text)
    step_expected_rc = db.Column("expected_rc", db.Integer)
    step_system_kwargs = db.Column("system_kwargs", db.JSON)
    target = db.Column(ScalarListType(str))
    created_on = db.Column(UtcDateTime(), nullable=False, default=get_now())
    step_action_type = db.Column("action_type", typos.Enum(ActionType))
    step_code = db.Column("code", db.Text)
    step_post_process = db.Column("post_process", db.Text)
    step_pre_process = db.Column("pre_process", db.Text)
    step_name = db.Column("name", db.String(40))

    orchestration = db.relationship("Orchestration", primaryjoin="Step.orchestration_id==Orchestration.id",
                                    back_populates="steps")
    action_template = db.relationship("ActionTemplate", primaryjoin="Step.action_template_id==ActionTemplate.id",
                                      backref="steps")

    parent_steps = db.relationship("Step", secondary="D_step_step",
                                   primaryjoin="D_step.c.id==D_step_step.c.step_id",
                                   secondaryjoin="D_step.c.id==D_step_step.c.parent_step_id",
                                   back_populates="children_steps")

    children_steps = db.relationship("Step", secondary="D_step_step",
                                     primaryjoin="D_step.c.id==D_step_step.c.parent_step_id",
                                     secondaryjoin="D_step.c.id==D_step_step.c.step_id",
                                     back_populates="parent_steps")

    def __init__(self, orchestration: 'Orchestration', undo: bool, action_template: 'ActionTemplate' = None,
                 action_type: ActionType = None, code: str = None, pre_process: str = None, post_process: str = None,
                 stop_on_error: bool = None, stop_undo_on_error: bool = None, undo_on_error: bool = None,
                 expected_stdout: t.Optional[str] = None,
                 expected_stderr: t.Optional[str] = None,
                 expected_rc: t.Optional[int] = None, parameters: t.Dict[str, t.Any] = None,
                 system_kwargs: t.Dict[str, t.Any] = None,
                 parent_steps: t.List['Step'] = None, children_steps: t.List['Step'] = None,
                 target: t.Union[str, t.Iterable[str]] = None, name=None, **kwargs):

        UUIDistributedEntityMixin.__init__(self, **kwargs)
        assert undo in (False, True)
        if action_template is not None:
            assert action_type is None
        else:
            assert action_type is not None
        self.undo = undo
        self.step_stop_on_error = stop_on_error if stop_on_error is not None else kwargs.pop('step_stop_on_error', None)
        self.step_stop_undo_on_error = stop_undo_on_error if stop_undo_on_error is not None else kwargs.pop(
            'step_stop_undo_on_error', None)
        self.step_undo_on_error = undo_on_error if undo_on_error is not None else kwargs.pop('step_undo_on_error', None)
        self.action_template = action_template
        self.step_action_type = action_type if action_type is not None else kwargs.pop('step_action_type', None)
        self.step_expected_stdout = expected_stdout if expected_stdout is not None else kwargs.pop(
            'step_expected_stdout', None)
        self.step_expected_stderr = expected_stderr if expected_stderr is not None else kwargs.pop(
            'step_expected_stderr', None)
        self.step_expected_rc = expected_rc if expected_rc is not None else kwargs.pop('step_expected_rc', None)
        self.step_parameters = parameters if parameters is not None else kwargs.pop('step_parameters', None) or {}
        self.step_system_kwargs = system_kwargs if system_kwargs is not None else kwargs.pop('step_system_kwargs',
                                                                                             None) or {}
        self.step_code = code if code is not None else kwargs.pop('step_code', None)
        self.step_post_process = post_process if post_process is not None else kwargs.pop('step_post_process', None)
        self.step_pre_process = pre_process if pre_process is not None else kwargs.pop('step_pre_process', None)
        self.orchestration = orchestration
        self.parent_steps = parent_steps or []
        self.children_steps = children_steps or []
        if self.undo is False:
            if target is None:
                self.target = ['all']
            else:
                self.target = [target] if isinstance(target, str) else (target if len(target) > 0 else ['all'])
        else:
            if target:
                raise errors.BaseError('target must not be set when creating an UNDO step')
        self.created_on = kwargs.get('created_on') or get_now()
        self.step_name = name if name is not None else kwargs.pop('step_name', None)

    @orm.reconstructor
    def init_on_load(self):
        self.parameters = self.parameters or {}
        self.system_kwargs = self.system_kwargs or {}

    @property
    def parents(self):
        return self.parent_steps

    @property
    def children(self):
        return self.children_steps

    @property
    def parent_undo_steps(self):
        return [s for s in self.parent_steps if s.undo == True]

    @property
    def children_undo_steps(self):
        return [s for s in self.children_steps if s.undo == True]

    @property
    def parent_do_steps(self):
        return [s for s in self.parent_steps if s.undo == False]

    @property
    def children_do_steps(self):
        return [s for s in self.children_steps if s.undo == False]

    @property
    def stop_on_error(self):
        return self.step_stop_on_error if self.step_stop_on_error is not None else self.orchestration.stop_on_error

    @stop_on_error.setter
    def stop_on_error(self, value):
        self.step_stop_on_error = value

    @property
    def stop_undo_on_error(self):
        return self.step_stop_undo_on_error if self.step_stop_undo_on_error is not None \
            else self.orchestration.stop_undo_on_error

    @stop_undo_on_error.setter
    def stop_undo_on_error(self, value):
        self.step_stop_undo_on_error = value

    @property
    def undo_on_error(self):
        if self.undo:
            return None
        return self.step_undo_on_error if self.step_undo_on_error is not None else self.orchestration.undo_on_error

    @undo_on_error.setter
    def undo_on_error(self, value):
        self.step_undo_on_error = value

    @property
    def parameters(self):
        if self.action_template:
            return dict(ChainMap(self.step_parameters, self.action_template.parameters))
        else:
            return dict(self.step_parameters)

    @parameters.setter
    def parameters(self, value):
        self.step_parameters = value

    @property
    def system_kwargs(self):
        if self.action_template:
            return dict(ChainMap(self.step_system_kwargs, self.action_template.system_kwargs))
        else:
            return dict(self.step_system_kwargs)

    @system_kwargs.setter
    def system_kwargs(self, value):
        self.step_system_kwargs = value

    @property
    def code(self):
        if self.step_code is None and self.action_template:
            return self.action_template.code
        else:
            return self.step_code

    @code.setter
    def code(self, value):
        self.step_code = value

    @property
    def action_type(self):
        if self.step_action_type is None and self.action_template:
            return self.action_template.action_type
        else:
            return self.step_action_type

    @action_type.setter
    def action_type(self, value):
        self.step_action_type = value

    @property
    def post_process(self):
        if self.step_post_process is None and self.action_template:
            return self.action_template.post_process
        else:
            return self.step_post_process

    @post_process.setter
    def post_process(self, value):
        self.step_post_process = value

    @property
    def pre_process(self):
        if self.step_pre_process is None and self.action_template:
            return self.action_template.pre_process
        else:
            return self.step_pre_process

    @pre_process.setter
    def pre_process(self, value):
        self.step_pre_process = value

    @property
    def expected_stdout(self):
        if self.step_expected_stdout is None and self.action_template:
            return self.action_template.expected_stdout
        else:
            return self.step_expected_stdout

    @expected_stdout.setter
    def expected_stdout(self, value):
        self.step_expected_stdout = value

    @property
    def expected_stderr(self):
        if self.step_expected_stderr is None and self.action_template:
            return self.action_template.expected_stderr
        else:
            return self.step_expected_stderr

    @expected_stderr.setter
    def expected_stderr(self, value):
        if value == self.action_template.expected_stderr:
            self.step_expected_stderr = None
        else:
            self.step_expected_stderr = value

    @property
    def expected_rc(self):
        if self.step_expected_rc is None and self.action_template:
            return self.action_template.expected_rc
        else:
            return self.step_expected_rc

    @expected_rc.setter
    def expected_rc(self, value):
        if self.action_template and value == self.action_template.expected_rc:
            self.step_expected_rc = None
        else:
            self.step_expected_rc = value

    @property
    def name(self):
        if self.step_name is None and self.action_template:
            return str(self.action_template)
        else:
            return self.step_name

    @name.setter
    def name(self, value):
        self.step_name = value

    @property
    def fetched_parameters(self):
        return set(re.findall(r'\(\?P<(\w+)>', self.regexp_fetch, flags=re.MULTILINE))

    @property
    def user_parameters(self) -> t.Set['str']:
        code_params = set(self.code_parameters)

        defined_params = set(self.parameters.keys())
        params_from_param_value = set()
        for v in self.parameters.values():
            params_from_param_value.union(re.findall(r'\{\{\s*([\.\w]+)\s*\}\}', str(v), flags=re.MULTILINE))

        return code_params.union(params_from_param_value).difference(defined_params)

    @property
    def code_parameters(self):
        return infer(self.code).keys()

    def eq_imp(self, other):
        """
        two steps are equal if they execute the same code with the same parameters even if they are from different
        orchestrations or they are in the same orchestration with different positions

        Parameters
        ----------
        other: Step

        Returns
        -------
        result: bool

        Notes
        -----
        _id and _orchestration are not compared
        """
        if isinstance(other, self.__class__):
            return all([self.undo == other.undo, self.parameters == other.parameters,
                        self.stop_on_error == other.stop_on_error,
                        self.stop_undo_on_error == other.stop_undo_on_error,
                        self.undo_on_error == other.undo_on_error,
                        self.expected_stdout == other.expected_stdout,
                        self.expected_stderr == other.expected_stderr,
                        self.expected_rc == other.expected_rc,
                        self.system_kwargs == other.system_kwargs,
                        self.code == other.code,
                        self.post_process == other.post_process,
                        self.pre_process == other.pre_process,
                        self.action_type == other.action_type,
                        ])
        else:
            raise NotImplemented

    def __str__(self):
        return self.name

    def __repr__(self):
        return ('Undo ' if self.undo else '') + self.__class__.__name__ + ' ' + str(getattr(self, 'id', ''))

    def _add_parents(self, parents):
        for step in parents:
            if not step in self.parent_steps:
                self.parent_steps.append(step)

    def _remove_parents(self, parents):
        for step in parents:
            if step in self.parent_steps:
                self.parent_steps.remove(step)

    def _add_children(self, children):
        for step in children:
            if not step in self.children_steps:
                self.children_steps.append(step)

    def _remove_children(self, children):
        for step in children:
            if step in self.children_steps:
                self.children_steps.remove(step)

    def to_json(self, add_action=False):
        data = super().to_json()
        if getattr(self.orchestration, 'id', None):
            data.update(orchestration_id=str(self.orchestration.id))
        if getattr(self.action_template, 'id', None):
            if add_action:
                data.update(action_template=self.action_template.to_json())
            else:
                data.update(action_template_id=str(self.action_template.id))
        data.update(undo=self.undo)
        data.update(stop_on_error=self.step_stop_on_error) if self.step_stop_on_error is not None else None
        data.update(
            stop_undo_on_error=self.step_stop_undo_on_error) if self.step_stop_undo_on_error is not None else None
        data.update(undo_on_error=self.step_undo_on_error) if self.step_undo_on_error is not None else None
        data.update(parameters=self.step_parameters) if self.step_parameters is not None else None
        data.update(expected_stdout=self.step_expected_stdout) if self.step_expected_stdout is not None else None
        data.update(expected_stderr=self.step_expected_stderr) if self.step_expected_stderr is not None else None
        data.update(expected_rc=self.step_expected_rc) if self.step_expected_rc is not None else None
        data.update(system_kwargs=self.step_system_kwargs) if self.step_system_kwargs is not None else None
        data.update(parent_step_ids=[str(step.id) for step in self.parents])
        data.update(code=self.step_code) if self.step_code is not None else None
        data.update(action_type=self.step_action_type.name) if self.step_action_type is not None else None
        data.update(post_process=self.step_post_process) if self.step_post_process is not None else None
        data.update(pre_process=self.step_pre_process) if self.step_pre_process is not None else None
        data.update(created_on=self.created_on.strftime(defaults.DATETIME_FORMAT))

        return data

    @classmethod
    def from_json(cls, kwargs):
        from dm.domain.entities import ActionTemplate, Orchestration
        kwargs = dict(kwargs)
        if 'orchestration_id' in kwargs:
            ident = kwargs.pop('orchestration_id')
            kwargs['orchestration'] = Orchestration.query.get(ident)
            if kwargs['orchestration'] is None:
                raise errors.EntityNotFound('Orchestration', ident=ident)
        if 'action_template_id' in kwargs:
            ident = kwargs.pop('action_template_id')
            kwargs['action_template'] = ActionTemplate.query.get(ident)
            if kwargs['action_template'] is None:
                raise errors.EntityNotFound('ActionTemplate', ident=ident)
        if 'action_type' in kwargs:
            kwargs['action_type'] = ActionType[kwargs.pop('action_type')]
        if 'created_on' in kwargs:
            kwargs['created_on'] = datetime.datetime.strptime(kwargs['created_on'], defaults.DATETIME_FORMAT)
        kwargs['parent_steps'] = []
        for parent_step_id in kwargs.pop('parent_step_ids', []):
            ps = Step.query.get(parent_step_id)
            if ps:
                kwargs['parent_steps'].append(ps)
            else:
                raise errors.EntityNotFound('Step', parent_step_id)
        return super().from_json(kwargs)




from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Orchestration, ActionTemplate, ActionType
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.web import create_app, db
from tests.helpers import TestCaseLockBypass


class Test(TestCaseLockBypass):


    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        self.auth = HTTPBearerAuth(create_access_token('00000000-0000-0000-0000-000000000001'))
        set_initial()
        self.at = ActionTemplate(name='action_name', version=1, action_type=ActionType.SHELL, code='')
        self.o = Orchestration(name='name', version=1)
        self.s1 = self.o.add_step(undo=False, action_template=self.at)
        self.s2 = self.o.add_step(undo=False, action_template=self.at, parents=[self.s1])
        self.s3 = self.o.add_step(undo=False, action_template=self.at, parents=[self.s2])
        db.session.add_all([self.at, self.o])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def test_get(self):
        resp = self.client.get(url_for('api_1_0.steprelationshipparents', step_id=str(self.s1.id)),
                               headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(dict(parent_step_ids=[]), resp.get_json())

        resp = self.client.get(url_for('api_1_0.steprelationshipparents', step_id=str(self.s2.id)),
                               headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(dict(parent_step_ids=[str(self.s1.id)]), resp.get_json())

    def test_patch(self):
        resp = self.client.patch(url_for('api_1_0.steprelationshipparents', step_id=str(self.s3.id)),
                                 json={'parent_step_ids': [str(self.s1.id)]},
                                 headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(dict(parent_step_ids=[str(self.s1.id)]), resp.get_json())
        self.assertListEqual([self.s1], self.s3.parents)

    def test_post(self):
        resp = self.client.post(url_for('api_1_0.steprelationshipparents', step_id=str(self.s3.id)),
                                json={'parent_step_ids': [str(self.s1.id)]},
                                headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        expected = [str(self.s2.id), str(self.s1.id)]
        expected.sort()
        actual = resp.get_json()['parent_step_ids']
        actual.sort()
        self.assertListEqual(expected, actual)
        self.assertIn(self.s1, self.s3.parents)
        self.assertIn(self.s2, self.s3.parents)

    def test_delete(self):
        resp = self.client.delete(url_for('api_1_0.steprelationshipparents', step_id=str(self.s3.id)),
                                  json={'parent_step_ids': [str(self.s2.id)]},
                                  headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(dict(parent_step_ids=[]), resp.get_json())
        self.assertEqual(0, len(self.s3.parents))



import datetime
from unittest import TestCase, mock

from dm.utils.async_operator import AsyncOperator

from dm.use_cases.deployment import UndoCommand, StepExecution, Command, CompositeCommand
from dm.utils.helpers import get_now


class PickableMock(mock.Mock):
    def __reduce__(self):
        return mock.Mock, ()


class TestCompositeCommand(TestCase):
    maxDiff = None

    @classmethod
    def setUpClass(cls) -> None:
        super().setUpClass()
        cls.ao = AsyncOperator()
        cls.ao.start()

    @classmethod
    def tearDownClass(cls) -> None:
        cls.ao.stop()
        super().tearDownClass()

    def test_force_all(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=get_now(),
                                                         end_time=get_now() + datetime.timedelta(
                                                             5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=get_now(),
                                                          end_time=get_now() + datetime.timedelta(
                                                              5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_error, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        cc1 = CompositeCommand({uc1: [uc2], uc2: [uc3]}, force_all=True, id_=1, async_operator=self.ao)

        res = cc1.invoke()

        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

    def test_force_all_undo(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=get_now(),
                                                         end_time=get_now() + datetime.timedelta(
                                                             5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=get_now(),
                                                          end_time=get_now() + datetime.timedelta(
                                                              5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_error, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_succ, undo_implementation=uc2, id_=2)
        c3 = Command(implementation=mocked_imp_error, undo_implementation=uc3, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, force_all=True, id_=1, async_operator=self.ao)

        res = cc1.invoke()

        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(2, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

    def test_undo_on_error(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=get_now(),
                                                         end_time=get_now() + datetime.timedelta(
                                                             5 / (24 * 60 * 60)))

        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=get_now(),
                                                          end_time=get_now() + datetime.timedelta(
                                                              5 / (24 * 60 * 60)))

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_error, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=uc1, id_=1)
        c2 = Command(implementation=mocked_imp_succ, undo_implementation=uc2, id_=2)
        c3 = Command(implementation=mocked_imp_error, undo_implementation=uc3, undo_on_error=True, id_=3)

        cc1 = CompositeCommand({c1: [c2], c2: [c3]}, force_all=True, id_=1, async_operator=self.ao)

        res = cc1.invoke()

        self.assertFalse(res)

        res = cc1.undo()

        self.assertEqual(3, mocked_imp_succ.execute.call_count)
        self.assertEqual(2, mocked_imp_error.execute.call_count)
        self.assertFalse(res)

    def test_composite_command_error(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=start_time,
                                                         end_time=end_time)
        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=start_time,
                                                          end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)

        ccu1 = CompositeCommand({uc1: []}, async_operator=self.ao)
        ccu2 = CompositeCommand({uc2: [uc3], uc3: [uc4]}, async_operator=self.ao)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=ccu1, id_=5)
        c2 = Command(implementation=mocked_imp_error, undo_implementation=ccu2, id_=6)

        cc = CompositeCommand({c1: [c2], c2: []}, async_operator=self.ao)

        res = cc.invoke()

        self.assertDictEqual({5: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              6: StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time)}
                             , cc.execution)

        self.assertEqual(False, res)
        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

        res = cc.undo()

        self.assertDictEqual({5: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              6: StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              1: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time)}
                             , cc.execution)

        self.assertEqual(True, res)
        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

    def test_composite_command_error2(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=start_time,
                                                         end_time=end_time)
        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=start_time,
                                                          end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)
        uc5 = UndoCommand(implementation=mocked_imp_succ, id_=5)

        ccu1 = CompositeCommand({uc1: []}, async_operator=self.ao)
        ccu2 = CompositeCommand({uc2: [uc3], uc3: [uc4]}, async_operator=self.ao)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=ccu1, id_=6)
        c2 = Command(implementation=mocked_imp_error, undo_implementation=ccu2, id_=7)
        c3 = Command(implementation=mocked_imp_succ, undo_implementation=uc5, id_=8)

        cc = CompositeCommand({c1: [c2], c2: [c3]}, async_operator=self.ao)

        res = cc.invoke()

        self.assertEqual(False, res)
        self.assertEqual(1, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

        res = cc.undo()

        self.assertEqual(True, res)
        self.assertEqual(2, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

    def test_composite_command_error3(self):
        mocked_imp_succ = PickableMock()
        mocked_imp_error = PickableMock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=start_time,
                                                         end_time=end_time)
        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=start_time,
                                                          end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_error, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)
        uc5 = UndoCommand(implementation=mocked_imp_succ, id_=5)

        ccu1 = CompositeCommand({uc1: []}, async_operator=self.ao)
        ccu2 = CompositeCommand({uc2: [uc3, uc4]}, async_operator=self.ao)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=ccu1, id_=6)
        c2 = Command(implementation=mocked_imp_succ, undo_implementation=ccu2, id_=7)
        c3 = Command(implementation=mocked_imp_succ, undo_implementation=uc5, id_=8)

        cc = CompositeCommand({c1: [c2], c2: [c3]}, async_operator=self.ao)

        res = cc.invoke()

        self.assertTrue(res)
        self.assertEqual(3, mocked_imp_succ.execute.call_count)
        self.assertEqual(0, mocked_imp_error.execute.call_count)

        res = cc.undo()

        self.assertFalse(res)
        self.assertEqual(6, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)

        self.assertDictEqual({2: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              3: StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              4: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              5: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              6: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              7: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              8: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time)
                              }
                             , cc.execution)

    def test_composite_command_success(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=start_time,
                                                         end_time=end_time)
        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=start_time,
                                                          end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=uc1, id_=5)
        c2 = Command(implementation=mocked_imp_succ, undo_implementation=uc2, id_=6)
        c3 = Command(implementation=mocked_imp_succ, undo_implementation=uc3, id_=7)
        c4 = Command(implementation=mocked_imp_succ, undo_implementation=uc4, id_=8)

        cc = CompositeCommand({c1: [c2, c3, c4]}, async_operator=self.ao)

        res = cc.invoke()

        self.assertDictEqual({5: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              6: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              7: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              8: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              }
                             , cc.execution)
        self.assertEqual(True, res)
        self.assertEqual(4, mocked_imp_succ.execute.call_count)
        self.assertEqual(0, mocked_imp_error.execute.call_count)

    def test_composite_command_success(self):
        mocked_imp_succ = mock.Mock()
        mocked_imp_error = mock.Mock()

        start_time = get_now()
        end_time = get_now() + datetime.timedelta(5 / (24 * 60 * 60))

        mocked_imp_succ.execute.return_value = StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                                         start_time=start_time,
                                                         end_time=end_time)
        mocked_imp_error.execute.return_value = StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                                          start_time=start_time,
                                                          end_time=end_time)

        uc1 = UndoCommand(implementation=mocked_imp_succ, id_=1)
        uc2 = UndoCommand(implementation=mocked_imp_succ, id_=2)
        uc3 = UndoCommand(implementation=mocked_imp_succ, id_=3)
        uc4 = UndoCommand(implementation=mocked_imp_succ, id_=4)

        c1 = Command(implementation=mocked_imp_succ, undo_implementation=uc1, id_=5)
        c2 = Command(implementation=mocked_imp_succ, undo_implementation=uc2, id_=6)
        c3 = Command(implementation=mocked_imp_error, undo_implementation=uc3, id_=7)
        c4 = Command(implementation=mocked_imp_succ, undo_implementation=uc4, id_=8)

        cc = CompositeCommand({c1: [c2, c3, c4]}, async_operator=self.ao)

        res = cc.invoke()

        self.assertDictEqual({5: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              6: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              7: StepExecution(success=False, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              8: StepExecution(success=True, stdout='stdout', stderr='stderr', rc=0,
                                           start_time=start_time,
                                           end_time=end_time),
                              }
                             , cc.execution)
        self.assertEqual(False, res)
        self.assertEqual(3, mocked_imp_succ.execute.call_count)
        self.assertEqual(1, mocked_imp_error.execute.call_count)



import argparse
import json
import os
import shlex
from pprint import pprint

from click import prompt
from prompt_toolkit.auto_suggest import AutoSuggestFromHistory
from prompt_toolkit.history import InMemoryHistory
from prompt_toolkit.lexers import PygmentsLexer
from pygments.lexers.python import PythonLexer
from pygments.lexers.shell import BashLexer
from pygments.lexers.special import TextLexer

import dm.dshell.network as ntwrk
from dm.domain.entities import ActionType
from dm.dshell.argparse_raise import ArgumentParserRaise
from dm.dshell.helpers import get_history, exit_dshell
from dm.dshell.prompts.utils import prompt_parameter
from dm.dshell.validators import ChoiceValidator, IntValidator, JSONValidator

parser = ArgumentParserRaise(allow_abbrev=False, prog='')
subparser = parser.add_subparsers(dest='cmd')
preview_action = subparser.add_parser('preview')
preview_action.set_defaults(func=lambda x: pprint(x))
set_action = subparser.add_parser('set')
set_action.add_argument('parameter')
submit_action = subparser.add_parser('submit')
dump_action = subparser.add_parser('dump')
dump_action.add_argument('file')
exit_action = subparser.add_parser('exit')


def code_lexer(action_type):
    if action_type == 'SHELL':
        lexer = PygmentsLexer(BashLexer)
    elif action_type == 'PYTHON':
        lexer = PygmentsLexer(PythonLexer)
    else:
        lexer = PygmentsLexer(TextLexer)
    return lexer


form = {
    "name": dict(history=InMemoryHistory()),
    "version": dict(history=InMemoryHistory()),
    "action_type": dict(validator=ChoiceValidator([at.name for at in ActionType if at.name != 'NATIVE']),
                        history=InMemoryHistory()),
    "code": dict(multiline=True, lexer=code_lexer, history=InMemoryHistory()),
    "expected_stdout": dict(multiline=True, history=InMemoryHistory()),
    "expected_stderr": dict(multiline=True, history=InMemoryHistory()),
    "expected_rc": dict(validator=IntValidator(), history=InMemoryHistory()),
    "parameters": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), validator=JSONValidator(),
                       history=InMemoryHistory()),
    "system_kwargs": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), validator=JSONValidator(),
                          history=InMemoryHistory()),
    "pre_process": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), history=InMemoryHistory()),
    "post_process": dict(multiline=True, lexer=PygmentsLexer(PythonLexer), history=InMemoryHistory()),
}

history = None

entity_name = os.path.basename(__file__).rstrip('.py')


def subprompt(entity, changed=False, ask_all=False, parent_prompt=None):
    global history
    if history is None:
        history = get_history(entity_name, InMemoryHistory())

    if ask_all:
        for k, v in form.items():
            if k not in entity:
                try:
                    if prompt_parameter(k, entity, entity_name, form, parent_prompt):
                        changed = True
                except KeyboardInterrupt:
                    entity[k] = None
                except EOFError:
                    exit_dshell()

    while True:
        try:
            text = prompt(f"{parent_prompt}{entity_name}('{entity['name']}')> ", history=history,
                          auto_suggest=AutoSuggestFromHistory())
        except KeyboardInterrupt:
            continue
        except EOFError:
            exit_dshell(rc=1)
        try:
            namespace = parser.parse_args(shlex.split(text))
        except (ValueError, argparse.ArgumentError) as e:
            print(e)
            continue
        except SystemExit:
            continue

        if namespace.cmd == 'preview':
            pprint(entity)
        elif namespace.cmd == 'set':
            if namespace.parameter not in form.keys():
                print("Not a valid parameter. Available: " + ', '.join(form.keys()))
            else:
                try:
                    if prompt_parameter(namespace.parameter, entity, form, f"{parent_prompt}{entity_name}('{entity['name']}')"):
                        changed = True
                except KeyboardInterrupt:
                    continue  # Control-C pressed. Try again.
                except EOFError:
                    exit_dshell(rc=1)
        elif namespace.cmd == 'submit':
            if 'id' in entity:
                resp = ntwrk.patch(f'api_1_0.{entity_name.replace("_", "")}resource', json=entity)
            else:
                resp = ntwrk.post(f'api_1_0.{entity_name.replace("_", "")}list', json=entity)

            if resp.code and 200 <= resp.code <= 299:
                if resp.msg:
                    pprint(resp.msg)
                break
            else:
                pprint(resp.msg if resp.code is not None else str(resp.exception) if str(
                    resp.exception) else resp.exception.__class__.__name__)
        elif namespace.cmd == 'dump':
            with open(namespace.file, 'w') as dumpfile:
                json.dump(entity, dumpfile, indent=4)
            changed = False
        elif namespace.cmd == 'exit':
            if changed:
                text = ''
                while text.lower().strip() not in ('y', 'n'):
                    text = prompt('If you exit you will lose the changes. Do you want to continue? (y/n): ')
                if text.lower().strip() == 'y':
                    break
            else:
                break



import sys as _sys
from asyncio import *

if _sys.version_info < (3, 7):
    def _set_task_name(task, name):
        if name is not None:
            try:
                set_name = task.set_name
            except AttributeError:
                pass
            else:
                set_name(name)


    def create_task(coro, *, name=None):
        """Schedule the execution of a coroutine object in a spawn task.

        Return a Task object.
        """
        loop = get_event_loop()
        task = loop.create_task(coro)
        _set_task_name(task, name)
        return task


    def run(aw, *, debug=False):
        # Emulate asyncio.run() on older versions
        loop = new_event_loop()
        set_event_loop(loop)
        loop.set_debug(debug)
        try:
            return loop.run_until_complete(aw)
        finally:
            loop.close()
            set_event_loop(None)



import datetime
import json
import logging
import os
import typing as t
import uuid
from collections import namedtuple

import aiohttp
import requests
from bs4 import BeautifulSoup
from flask import current_app
from pkg_resources import parse_version

from dm import __version__ as dm_version, defaults
from dm.domain.entities import Server, Catalog, Route
from dm.network.low_level import check_host
from dm.use_cases.helpers import get_auth_root
from dm.use_cases.use_cases import run_elevator, get_software, upgrade_catalog_from_server
from dm.utils import asyncio
from dm.utils.asyncio import create_task
from dm.utils.helpers import get_filename_from_cd, convert
from dm.web import db, errors
from dm.web.decorators import run_as
from dm.web.network import async_get, ping, get, Response

logger = logging.getLogger('dm.background')
routing_logger = logging.getLogger('dm.background.routing')
catalog_logger = logging.getLogger('dm.background.catalog')
upgrader_logger = logging.getLogger('dm.background.upgrader')



@run_as('root')
def process_get_new_version_from_gogs(app=None):
    """
    checks if new version in repo

    Parameters
    ----------
    app:
        app to load the context
    timeout_wait_transfer:
        timeout waiting tranfer file to end.
    refresh_interval:
        time period to check if tranfer ended. Normally, used for test purposes

    Returns
    -------

    """
    upgrader_logger.info('Starting Upgrade Process')
    base_url = os.environ.get('GIT_REPO') \
               or current_app.config.get('GIT_REPO') \
               or 'https://ca355c55-0ab0-4882-93fa-331bcc4d45bd.pub.cloud.scaleway.com:3000'
    releases_uri = '/dimensigon/dimensigon/releases'
    try:
        r = requests.get(base_url + releases_uri, verify=current_app.config['SSL_VERIFY'], timeout=10)
    except (requests.exceptions.Timeout, requests.exceptions.ConnectionError) as e:
        r = None
        upgrader_logger.info('Unable to contact to main repo')

    # get new versions from repo
    if r and r.status_code == 200:
        # get current software
        gogs_versions = {}

        html_content = r.text
        soup = BeautifulSoup(html_content, 'html.parser')
        for li in soup.find(id='release-list').find_all('li'):
            version = li.h4.a.get_text(strip=True)
            uris = [a.attrs['href'] for a in li.find('div', class_='download').find_all('a') if
                    a.attrs['href'].endswith('tar.gz')]
            if len(uris) > 0:
                gogs_versions.update({parse_version(version): uris[0]})
        current_version = parse_version(dm_version)
        new_versions = [gogs_ver for gogs_ver in gogs_versions if gogs_ver > current_version]

        if new_versions:
            new_version = max(new_versions)
            upgrader_logger.info(f"Downloading version {new_version} from outside world")

            r = requests.get(base_url + gogs_versions[new_version],
                             verify=current_app.config['SSL_VERIFY'])
            filename = get_filename_from_cd(
                r.headers.get(
                    'content-disposition')) or f"dimensigon-{gogs_versions[new_version].rsplit('/', 1)[-1]}"
            os.makedirs(os.path.join(current_app.config['SOFTWARE_REPO'], 'dimensigon'), exist_ok=True)
            file = os.path.join(current_app.config['SOFTWARE_REPO'], 'dimensigon', filename)
            try:
                open(file, 'wb').write(r.content)
            except Exception as e:
                upgrader_logger.exception(f"Unable to save {file}")
            else:
                run_elevator(file, new_version, upgrader_logger)
    else:
        upgrader_logger.debug(f"No version to upgrade")


TempRoute = namedtuple('TempRoute', ['proxy_server', 'gate', 'cost'])


def update_table_routing_cost(discover_new_neighbours=False, check_current_neighbours=False, retries=2, timeout=10) -> \
        t.Dict[
            Server, TempRoute]:
    """Gets route tables of all neighbours and updates its own table based on jump weights.
    Needs a Flask App Context to run.

    Parameters
    ----------
    discover_new_neighbours:
        tries to discover new neighbours
    check_current_neighbours:
        checks if current neighbours are still neighoburs
    retries:
        number of times it will try to reach destination
    timeout:
        time in seconds to stop waiting for connection

    Returns
    -------
    None
    """
    # get all neighbours
    temp_table_routes: t.Dict[uuid.UUID, t.List[TempRoute]] = {}
    changed_routes: t.Dict[Server, TempRoute] = {}
    me = Server.get_current()
    not_neighbours = Server.get_not_neighbours()
    routing_logger.debug('Updating routing table')
    not_neighbours_anymore = []
    if check_current_neighbours:
        neighbours = Server.get_neighbours()
        routing_logger.debug(
            f"Checking current neighbours: " + ', '.join([str(s) for s in neighbours]))
        for server in neighbours:
            route = server.route
            cost, time = ping(server, me, retries=retries, timeout=timeout)
            if cost is None:
                default_gate = server.route.gate
                temp = [None, None, None]
                # try another gate
                for gate in server.gates:
                    # check not to connect with localhost gate from current node and not to check already checked gate
                    if gate != default_gate and ((gate.ip and not gate.ip.is_loopback) or (
                            gate.dns and gate.dns != 'localhost')):
                        if check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries, delay=0.2,
                                      timeout=timeout):
                            temp[0] = None
                            temp[1] = gate
                            temp[2] = 0
                            break
                if temp[2] is None:
                    not_neighbours_anymore.append(server)
                if route.proxy_server != temp[0] or route.gate != temp[1] or route.cost != temp[2]:
                    changed_routes[server] = TempRoute(*temp)
                    route.proxy_server, route.gate, route.cost = temp

            # try:
            #     requests.get(server.url('root.healthcheck'), timeout=0.5,
            #                  verify=False)
            # except (requests.exceptions.ConnectTimeout, TimeoutError):
            #     # TODO: handle when a neighobur is not a neighbour anymore
            #     server.cost = None
            # else:
            #     server.cost = 0
            #     server.gateway = None
    if len(not_neighbours_anymore) > 0:
        routing_logger.info(
            f"Lost direct connection to the following nodes: " + ', '.join([str(s) for s in not_neighbours_anymore]))

    new_neighbours = []
    if discover_new_neighbours:
        routing_logger.debug(
            f"Checking new neighbours: " + ', '.join([str(s) for s in not_neighbours]))
        for server in not_neighbours:
            for gate in server.gates:
                if (gate.ip and not gate.ip.is_loopback) or (gate.dns and gate.dns != 'localhost'):
                    if check_host(host=gate.dns or str(gate.ip), port=gate.port, retry=retries, delay=1,
                                  timeout=timeout):

                        if server.route:
                            server.route.gate = gate
                            server.route.proxy_server = None
                            server.route.cost = 0
                            new_neighbours.append(server)
                        else:
                            r = Route(destination=server, proxy_server=None, gate=gate, cost=0)
                            db.session.add(r)
                        changed_routes[server] = TempRoute(None, server.route.gate, server.route.cost)
                        break
        if new_neighbours:
            routing_logger.info(f'New neighbours found: ' + ', '.join([str(s) for s in new_neighbours]))

    pool_responses = []
    neighoburs = Server.get_neighbours()
    if new_neighbours or not_neighbours_anymore:
        routing_logger.info(f"New Neighbour list {', '.join([str(s) for s in neighoburs])}")

    for server in neighoburs:
        pool_responses.append(get(server, 'api_1_0.routes', auth=get_auth_root(), timeout=10))

    for resp in pool_responses:
        if resp.code == 200:

            likely_proxy_server_entity = db.session.query(Server).get(resp.msg.get('server_id'))
            routing_logger.debug(
                f"route list got from server {likely_proxy_server_entity}: {json.dumps(resp.msg['route_list'], indent=4)}")

            for route_json in resp.msg['route_list']:
                route_json = convert(route_json)
                if route_json.destination_id != me.id \
                        and route_json.proxy_server_id != me.id \
                        and route_json.gate_id not in [g.id for g in me.gates]:
                    if route_json.destination_id not in temp_table_routes:
                        temp_table_routes.update({route_json.destination_id: []})
                    if route_json.cost is not None:
                        route_json.cost += 1
                        route_json.proxy_server_id = likely_proxy_server_entity.id
                        route_json.gate_id = None
                        temp_table_routes[route_json.destination_id].append(
                            TempRoute(likely_proxy_server_entity.id, None, route_json.cost))
                    elif route_json.cost is None:
                        # remove a routing if gateway cannot reach the destination
                        temp_table_routes[route_json.destination_id].append(
                            TempRoute(route_json.proxy_server_id, None, None))
        else:
            s = neighoburs[pool_responses.index(resp)]
            routing_logger.error(f"Error while connecting with {s}. Error: {resp[1]}, {resp[0]}")

    # Select new routes based on neighbour routes
    MAX_COST = 9999999
    neighbour_ids = [s.id for s in Server.get_neighbours()]
    for destination_id in filter(lambda s: s not in neighbour_ids, temp_table_routes.keys()):
        route = db.session.query(Route).filter_by(destination_id=destination_id).one_or_none()
        if not route:
            server = Server.query.get(destination_id)
            if not server:
                continue
            else:
                route = Route(destination=server)
        temp_table_routes[destination_id].sort(key=lambda x: x.cost or MAX_COST)
        if len(temp_table_routes[destination_id]) > 0:
            min_route = temp_table_routes[destination_id][0]
            proxy_server = db.session.query(Server).get(min_route.proxy_server)
            cost = min_route.cost
            if route.proxy_server != proxy_server or route.cost != cost:
                route.proxy_server = proxy_server
                route.gate = None
                route.cost = cost
                changed_routes[route.destination] = TempRoute(route.proxy_server,
                                                              route.gate,
                                                              route.cost)
                db.session.add(route)
                break

    data = {}
    for server, temp_route in changed_routes.items():
        data.update({str(server): {'proxy_server': str(temp_route.proxy_server), 'gate': str(temp_route.gate),
                                   'cost': str(temp_route.cost)}})
    routing_logger.debug(f'Changed routes from neighbours: {json.dumps(data, indent=4)}')
    return changed_routes


async def _get_neighbour_catalog_data_mark() -> t.Dict[Server, Response]:
    server_responses = {}
    servers = Server.get_neighbours()
    catalog_logger.debug(f"Neighbour servers to check: {[s.name for s in servers]}")
    auth = get_auth_root()

    async with aiohttp.ClientSession(connector=aiohttp.TCPConnector(
            ssl=current_app.config['SSL_VERIFY'])) as session:
        for server in servers:
            server_responses[server] = create_task(async_get(server, 'root.healthcheck', session=session,
                                                             auth=auth))

        for server, future in server_responses.items():
            server_responses[server] = await future
            if server_responses[server].code == 200:
                id_response = server_responses[server].msg.get('server', {}).get('id', '')
                if id_response and str(server.id) != id_response:
                    server_responses[server].exception = errors.HealthCheckMismatch(
                        expected={'id': str(server.id), 'name': server.name},
                        actual=server_responses[server].msg.get('server', {}))
                    server_responses[server].msg = None
                    server_responses[server].code = None

            catalog_logger.debug(
                f"Response from server {server.name}: {json.dumps(server_responses[server].to_dict(), indent=4)}")

    return server_responses


def upgrade_version(data: t.Dict[Server, Response]):
    mayor_version, mayor_server = None, None
    for server, response in data.items():
        if response.code == 200 and 'version' in response.msg:
            remote_version = parse_version(response.msg['version'])
            if remote_version > parse_version(dm_version):
                if mayor_version is None or mayor_version < remote_version:
                    mayor_version, mayor_server = remote_version, server
    if mayor_version:
        catalog_logger.info(f'Found mayor version on server {mayor_server}. Upgrading version first')
        file, v = get_software(mayor_server, get_auth_root())
        if file:
            run_elevator(file, mayor_version, catalog_logger)
            return True
    return False


def update_catalog(data: t.Dict[Server, Response]):
    reference_server = None
    catalog_ver = db.session.query(db.func.max(Catalog.last_modified_at)).scalar()
    if catalog_ver:
        for server, response in data.items():
            if response.code == 200 and 'catalog_version' in response.msg:
                new_catalog_ver = datetime.datetime.strptime(response.msg['catalog_version'],
                                                             defaults.DATEMARK_FORMAT)
                if new_catalog_ver > catalog_ver:
                    if response.msg['version'] == dm_version:
                        reference_server = server
                        catalog_ver = new_catalog_ver
                    else:
                        catalog_logger.debug(
                            f"Server {server} has different software version {response.msg['version']}")
            else:
                msg = f"Error while trying to get healthcheck from server {server.name}. "
                if response.code:
                    msg = msg + f"Response from server (code {response.code}): {response.msg}"
                else:
                    msg = msg + f"Exception: {response.code}"
                catalog_logger.warning(msg)
        if reference_server:
            catalog_logger.info(f"New catalog found from server {reference_server.name}: {catalog_ver}")
            upgrade_catalog_from_server(reference_server)
        else:
            catalog_logger.debug(f"No server with higher catalog found")


@run_as('root')
def process_catalog_route_table(app=None):
    # app will be used for the run_as decorator
    if update_table_routing_cost(discover_new_neighbours=True, check_current_neighbours=True):
        db.session.commit()
    catalog_logger.debug("Starting check catalog from neighbours")
    data = asyncio.run(_get_neighbour_catalog_data_mark())
    # check version upgrade before catalog upgrade to match database revision
    if not upgrade_version(data):
        update_catalog(data)
    db.session.commit()






from unittest import TestCase
from unittest.mock import patch

from flask import Flask

from dm.domain.entities import Server, Route
from dm.network.exceptions import NotValidMessage
from dm.utils.helpers import generate_dimension
from dm.web import db
from dm.web.decorators import securizer


class TestSecurizer(TestCase):
    def setUp(self):
        """Create and configure a new self.app instance for each test."""
        # create a temporary file to isolate the database for each test
        # create the self.app with common test config

        self.app = Flask(__name__)

        @self.app.route('/', methods=['GET', 'POST'])
        @securizer
        def hello():
            return {'msg': 'default response'}

        self.app.config['SECURIZER_PLAIN'] = True
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.client = self.app.test_client()
        db.init_app(self.app)
        db.create_all()
        self.d = generate_dimension('test')
        self.srv1 = Server(id='bbbbbbbb-1234-5678-1234-56781234bbb1', name='server1',
                           dns_or_ip='192.168.1.9', port=7123, _me=True)
        Route(self.srv1, cost=0)

        db.session.add_all([self.srv1, self.d])
        db.session.commit()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.web.decorators.url_for')
    @patch('dm.web.decorators.pack_msg')
    def test_get_securizer(self, mock_pack_msg, mock_url_for):
        # mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        mock_pack_msg.return_value = {'data': 'encrypted data'}
        mock_url_for.return_value = '/join'

        resp = self.client.get('/')

        self.assertEqual(200, resp.status_code)
        mock_pack_msg.assert_called_once_with(data={'msg': 'default response'})
        self.assertDictEqual({'data': 'encrypted data'}, resp.get_json())

        resp = self.client.get('/', headers={'D-Securizer': 'plain'})

        self.assertEqual(200, resp.status_code)
        mock_pack_msg.assert_called_once()
        self.assertDictEqual({'msg': 'default response'}, resp.get_json())

    @patch('dm.web.decorators.url_for')
    @patch('dm.web.decorators.pack_msg')
    def test_get_securizer_plain_not_allowed(self, mock_pack_msg, mock_url_for):
        # mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        mock_pack_msg.return_value = {'data': 'encrypted data'}
        mock_url_for.return_value = '/join'
        self.app.config['SECURIZER_PLAIN'] = False
        resp = self.client.get('/', headers={'D-Securizer': 'plain'})

        self.assertEqual(406, resp.status_code)
        mock_pack_msg.assert_not_called()
        self.assertDictEqual({'error': 'plain data is not allowed'}, resp.get_json())

    @patch('dm.web.decorators.url_for')
    @patch('dm.web.decorators.unpack_msg')
    @patch('dm.web.decorators.pack_msg')
    def test_post_securizer(self, mock_pack_msg, mock_unpack_msg, mock_url_for):
        # mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        mock_pack_msg.return_value = {'data': 'encrypted data'}
        mock_unpack_msg.return_value = {'data': 'decrypted data'}
        mock_url_for.return_value = '/join'
        resp = self.client.post('/', json={'data': 'post data'})

        self.assertEqual(200, resp.status_code)
        mock_unpack_msg.assert_called_once_with(data={'data': 'post data'})
        mock_pack_msg.assert_called_once_with(data={'msg': 'default response'})
        self.assertDictEqual({'data': 'encrypted data'}, resp.get_json())

        resp = self.client.post('/', json={'data': 'post data'}, headers={'D-Securizer': 'plain'})

        self.assertEqual(200, resp.status_code)
        mock_unpack_msg.assert_called_once()
        mock_pack_msg.assert_called_once()
        self.assertDictEqual({'msg': 'default response'}, resp.get_json())

    @patch('dm.web.decorators.url_for')
    @patch('dm.web.decorators.unpack_msg')
    @patch('dm.web.decorators.pack_msg')
    def test_post_securizer_plain_not_allowed(self, mock_pack_msg, mock_unpack_msg, mock_url_for):
        # mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')
        mock_pack_msg.return_value = {'data': 'encrypted data'}
        mock_url_for.return_value = '/join'
        self.app.config['SECURIZER_PLAIN'] = False
        resp = self.client.post('/', json={'data': 'post data'}, headers={'D-Securizer': 'plain'})


        self.assertEqual(406, resp.status_code)

        mock_unpack_msg.assert_not_called()
        mock_pack_msg.assert_not_called()
        self.assertDictEqual({'error': 'plain data is not allowed'}, resp.get_json())

    @patch('dm.web.decorators.url_for')
    @patch('dm.web.decorators.unpack_msg')
    @patch('dm.web.decorators.pack_msg')
    def test_post_securizer_validation_error(self, mock_pack_msg, mock_unpack_msg, mock_url_for):
        # mock_g.server = MagicMock(id='bbbbbbbb-1234-5678-1234-56781234bbb1')

        mock_url_for.return_value = '/join'
        mock_unpack_msg.side_effect = NotValidMessage("Message")

        resp = self.client.post('/', json={'data': 'post data'})


        self.assertEqual(400, resp.status_code)
        self.assertDictEqual({'error': 'Message',
                                'message': {'data': 'post data'}}, resp.get_json())

        mock_unpack_msg.assert_called_once_with(data={'data': 'post data'})
        mock_pack_msg.assert_not_called()


import re
import traceback
from functools import partial
from unittest import TestCase, mock

from aioresponses import aioresponses, CallbackResult
from flask import url_for
from flask_jwt_extended import create_access_token

from dm.domain.entities import Server, Route, Dimension, User
from dm.domain.entities.bootstrap import set_initial
from dm.network.auth import HTTPBearerAuth
from dm.utils.helpers import generate_dimension
from dm.web import create_app, db
from dm.web.api_1_0.urls.use_cases import wrap_sudo


class TestLaunchCommand(TestCase):
    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        # self.app.config['SECURIZER'] = True
        self.app_context = self.app.app_context()
        self.app_context.push()
        self.dim = generate_dimension('test')
        self.dim.current = True
        self.json_dim = self.dim.to_json()
        self.client = self.app.test_client()

        set_initial(server=False)
        User.set_initial()

        self.auth = HTTPBearerAuth(create_access_token(User.get_by_user('root').id))
        server = Server('node1', port=8000, me=True, granules='granule')
        db.session.add_all([server, self.dim])
        db.session.commit()

        # dump data
        self.json_node1 = server.to_json(add_gates=True)
        self.json_users = [u.to_json() for u in User.query.all()]

        self.app2 = create_app('test')
        # self.app2.config['SECURIZER'] = True
        self.client2 = self.app2.test_client()
        with self.app2.app_context():
            set_initial(server=False)
            me = Server('node2', port=8000, me=True, granules='granule')
            db.session.add(me)

            src_server = Server.from_json(self.json_node1)
            Route(src_server, cost=0)
            db.session.add(src_server)

            dim = Dimension.from_json(self.json_dim)
            dim.current = True
            db.session.add(dim)

            users = [User.from_json(ju) for ju in self.json_users]
            db.session.add_all(users)

            db.session.commit()

            # dump data
            self.json_node2 = me.to_json(add_gates=True)

        node2 = Server.from_json(self.json_node2)
        Route(node2, cost=0)
        db.session.add(node2)
        db.session.commit()

    def tearDown(self) -> None:
        with self.app2.app_context():
            db.session.remove()
            db.drop_all()

        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    def set_callbacks(self, m):
        def callback_client(method, client, url, **kwargs):
            kwargs.pop('allow_redirects')
            # passing headers as a workarround for https://github.com/pnuckowski/aioresponses/issues/111
            func = getattr(client, method.lower())
            try:
                r = func(url.path, headers=kwargs['headers'], json=kwargs['json'])
            except Exception as e:
                return CallbackResult(method.upper(), status=500, body=traceback.format_exc(), headers={})

            return CallbackResult(method.upper(), status=r.status_code, body=r.data, content_type=r.content_type,
                                  headers=r.headers)

        m.post(re.compile('https?://node2:.*'),
               callback=partial(callback_client, 'POST', self.app2.test_client()), repeat=True)

    @aioresponses()
    @mock.patch('dm.web.api_1_0.urls.use_cases.subprocess.Popen')
    def test_launch_command(self, m, mock_popen):
        self.maxDiff = None
        self.set_callbacks(m)

        popen_mock = mock.MagicMock()
        mock_popen.return_value = popen_mock
        popen_mock.communicate.return_value = ('output', '')
        type(popen_mock).returncode = mock.PropertyMock(return_value=0)

        resp = self.client.post(url_for('api_1_0.launch_command'),
                                json={"command": "ls -l", "hosts": 'all', 'timeout': 1},
                                headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual({self.json_node1['id']: {'stdout': 'output', 'stderr': '', 'returncode': 0},
                              self.json_node2['id']: {'stdout': 'output', 'stderr': '', 'returncode': 0},
                              },
                             resp.get_json())

        self.assertEqual(2, mock_popen.call_count)

        self.assertTupleEqual((wrap_sudo('root', ['ls', '-l']), ), mock_popen.call_args[0])

        resp = self.client.post(url_for('api_1_0.launch_command'),
                                json={"command": "ls -l", "hosts": [self.json_node2['id']], 'timeout': 1},
                                headers=self.auth.header)
        self.assertEqual(200, resp.status_code)
        self.assertDictEqual({self.json_node2['id']: {'stdout': 'output', 'stderr': '', 'returncode': 0},
                              },
                             resp.get_json())

        resp = self.client.post(url_for('api_1_0.launch_command'),
                                json={"command": "ls -l", "hosts": self.json_node1['id'], 'timeout': 1},
                                headers=self.auth.header)
        self.assertEqual(200, resp.status_code)
        self.assertDictEqual({self.json_node1['id']: {'stdout': 'output', 'stderr': '', 'returncode': 0},
                              },
                             resp.get_json())

    @aioresponses()
    @mock.patch('dm.web.api_1_0.urls.use_cases.subprocess.Popen')
    def test_launch_command_timeout(self, m, mock_popen):
        self.maxDiff = None
        self.set_callbacks(m)

        args = wrap_sudo('root', ['sleep', '10'])
        cmd = 'sleep 10'
        popen_mock = mock.MagicMock()
        mock_popen.return_value = popen_mock
        popen_mock.communicate.side_effect = [TimeoutError, ('', '')]
        type(popen_mock).returncode = mock.PropertyMock(return_value=0)

        resp = self.client.post(url_for('api_1_0.launch_command'),
                                json={"command": cmd, "hosts": 'node1', 'timeout': 1},
                                headers=self.auth.header)

        self.assertEqual(200, resp.status_code)
        self.assertDictEqual(
            {self.json_node1['id']: {'error': f"Command '{wrap_sudo('root', cmd)}' timed out after 1 seconds",
                                     'stdout': '', 'stderr': ''},
             },
            resp.get_json())



    def test_launch_command_rm_recursive(self):
        resp = self.client.post(url_for('api_1_0.launch_command'),
                                json={"command": "rm -fr /folder", "hosts": "all", 'timeout': 1},
                                headers=self.auth.header)

        self.assertEqual(403, resp.status_code)
        self.assertDictEqual({'error': 'rm with recursion is not allowed'},
                             resp.get_json())


import gc
import gzip
import io
import os
import shutil
import sys
import tempfile
import unittest
from unittest.mock import patch

from pyfakefs.fake_filesystem_unittest import TestCase

from dm.utils.pygtail import Pygtail

PY2 = sys.version_info[0] == 2


class PygtailTest(TestCase):
    # TODO:
    # - test for non-default offset file
    # - test for savelog and datext rotation schemes

    def setUp(self):
        self.setUpPyfakefs()
        self.test_lines = ["1\n", "2\n", "3\n", "4\n", "5\n"]
        self.test_str = ''.join(self.test_lines)
        self.logfile = tempfile.NamedTemporaryFile(delete=False)
        self.logfile.write(self.test_str.encode('utf-8'))
        self.logfile.close()
        self.addCleanup(self.tearDown)


    def append(self, str):
        # append the give string to the temp logfile
        with open(self.logfile.name, "ab") as fh:
            fh.write(str.encode())

    def copytruncate(self):
        shutil.copyfile(self.logfile.name, "%s.1" % self.logfile.name)
        fh = open(self.logfile.name, "wb")
        fh.close()

    def tearDown(self):
        gc.collect()
        filename = self.logfile.name
        self.logfile.close()
        for tmpfile in [filename, filename + ".offset", filename + ".1", filename + ".1.gz"]:
            if os.path.exists(tmpfile):
                try:
                    os.remove(tmpfile)
                except PermissionError:
                    pass


    def test_read(self):
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), self.test_str)

    def test_read_binary(self):
        pygtail = Pygtail(self.logfile.name, binary=True)
        self.assertEqual(pygtail.read(), self.test_str.encode())

    def test_read_encoded(self):
        self.test_read()
        new_line = 'çÃô\n'
        self.append(new_line)
        pygtail = Pygtail(self.logfile.name, encoding='ascii', errors='ignore')
        self.assertEqual(pygtail.read(), '\n')
        self.append(new_line)
        pygtail = Pygtail(self.logfile.name, encoding='utf-8')
        self.assertEqual(pygtail.read(), new_line)

    def test_errors(self):
        self.test_read()
        new_line = b'\x80abc\n'
        with open(self.logfile.name, "ab") as fh:
            fh.write(new_line)
            fh.close()
        pygtail = Pygtail(self.logfile.name, encoding='utf-8', errors='strict')

        with self.assertRaises(UnicodeDecodeError) as e:
            pygtail.read()

        pygtail = Pygtail(self.logfile.name, encoding='utf-8', errors='replace')

        self.assertEqual('\ufffdabc\n', pygtail.read())

    def test_readlines(self):
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.readlines(), self.test_lines)

    def test_readlines_max_lines(self):
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(["1\n", "2\n", "3\n", "4\n"], pygtail.readlines(4))

        self.assertEqual(["5\n"], pygtail.readlines(2))

    def test_readlines_max_lines_binary(self):
        pygtail = Pygtail(self.logfile.name, binary=True)
        self.assertEqual([b"1\n", b"2\n", b"3\n", b"4\n"], pygtail.readlines(4))

        self.assertEqual([b"5\n"], pygtail.readlines(2))

    def test_readlines_max_characters(self):
        import dm.utils.pygtail as py
        py.MAX_LINE_SIZE = 1

        pygtail = py.Pygtail(self.logfile.name)
        self.assertEqual(['1', '\n'], pygtail.readlines(2))

    @patch('dm.utils.pygtail.Pygtail.logger')
    def test_readlines_max_characters_full_line(self, mocked_logger):
        import dm.utils.pygtail as py
        py.MAX_LINE_SIZE = 1

        pygtail = py.Pygtail(self.logfile.name, full_lines=True)
        self.assertEqual([], pygtail.readlines(2))
        self.assertTrue(mocked_logger.warning.called)
        py.MAX_LINE_SIZE = 100

        self.assertEqual(self.test_lines, pygtail.readlines())

    def test_subsequent_read_with_no_new_data(self):
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), self.test_str)
        self.assertEqual(pygtail.read(), '')

    def test_subsequent_read_with_new_data(self):
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), self.test_str)
        new_lines = "4\n5\n"
        self.append(new_lines)
        new_pygtail = Pygtail(self.logfile.name)
        self.assertEqual(new_pygtail.read(), new_lines)

    def test_read_from_the_file_end(self):
        pygtail = Pygtail(self.logfile.name, read_from_end=True)
        self.assertEqual(pygtail.read(), '')
        new_lines = "4\n5\n"
        self.append(new_lines)
        new_pygtail = Pygtail(self.logfile.name, read_from_end=True)
        self.assertEqual(new_pygtail.read(), new_lines)

    def test_logrotate_without_delay_compress(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])

        # put content to gzip file
        gzip_handle = gzip.open("%s.1.gz" % self.logfile.name, 'wb')
        with open(self.logfile.name, 'rb') as logfile:
            gzip_handle.write(logfile.read())
        gzip_handle.close()

        with open(self.logfile.name, 'w'):
            # truncate file
            pass

        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    @unittest.skipIf(sys.platform.startswith('win'), "lock file mechanism on windows does not allow rename")
    def test_logrotate_with_delay_compress(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])
        os.rename(self.logfile.name, "%s.1" % self.logfile.name)
        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    @unittest.skipIf(sys.platform.startswith('win'), "lock file mechanism on windows does not allow rename")
    def test_logrotate_with_dateext_with_delaycompress(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])
        os.rename(self.logfile.name, "%s-20160616" % self.logfile.name)
        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    def test_logrotate_with_dateext_without_delaycompress(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])

        # put content to gzip file
        gzip_handle = gzip.open("%s-20160616.gz" % self.logfile.name, 'wb')
        with open(self.logfile.name, 'rb') as logfile:
            gzip_handle.write(logfile.read())
        gzip_handle.close()

        with open(self.logfile.name, 'w'):
            # truncate file
            pass

        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    @unittest.skipIf(sys.platform.startswith('win'), "lock file mechanism on windows does not allow rename")
    def test_logrotate_with_dateext2_with_delaycompress(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])
        os.rename(self.logfile.name, "%s-20160616-1466093571" % self.logfile.name)
        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    def test_logrotate_with_dateext2_without_delaycompress(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])

        # put content to gzip file
        gzip_handle = gzip.open("%s-20160616-1466093571.gz" % self.logfile.name, 'wb')
        with open(self.logfile.name, 'rb') as logfile:
            gzip_handle.write(logfile.read())
        gzip_handle.close()

        with open(self.logfile.name, 'w'):
            # truncate file
            pass

        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    @unittest.skipIf(sys.platform.startswith('win'), "lock file mechanism on windows does not allow rename")
    def test_timed_rotating_file_handler(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])
        os.rename(self.logfile.name, "%s.2016-06-16" % self.logfile.name)
        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name)
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    @unittest.skipIf(sys.platform.startswith('win'), "lock file mechanism on windows does not allow rename")
    def test_custom_rotating_file_handler_with_prepend(self):
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        self.append(new_lines[0])
        file_dir, rel_filename = os.path.split(self.logfile.name)
        os.rename(self.logfile.name, os.path.join(file_dir, "custom_log_pattern.%s" % rel_filename))
        self.append(new_lines[1])
        pygtail = Pygtail(self.logfile.name, log_patterns=["custom_log_pattern.%s"])
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    def test_copytruncate_off_smaller(self):
        self.test_readlines()
        self.copytruncate()
        new_lines = "4\n5\n"
        self.append(new_lines)

        old_stderr = sys.stderr
        sys.stderr = captured = io.BytesIO() if PY2 else io.StringIO()
        pygtail = Pygtail(self.logfile.name, copytruncate=False)
        captured_value = captured.getvalue()
        sys.stderr = old_stderr

        assert_class = self.assertRegex if sys.version_info >= (3, 1) else self.assertRegexpMatches
        assert_class(captured_value, r".*?\bWARN\b.*?\bshrank\b.*")
        self.assertEqual(pygtail.read(), '')

    def test_copytruncate_on_smaller(self):
        self.test_readlines()
        self.copytruncate()
        new_lines = "4\n5\n"
        self.append(new_lines)
        pygtail = Pygtail(self.logfile.name, copytruncate=True)
        self.assertEqual(pygtail.read(), new_lines)

    def _test_copytruncate_larger(self, onoff):
        self.test_readlines()
        self.copytruncate()
        self.append(self.test_str)
        new_lines = "4\n5\n"
        self.append(new_lines)
        pygtail = Pygtail(self.logfile.name, copytruncate=onoff)
        self.assertEqual(pygtail.read(), new_lines)

    def test_copytruncate_larger_off(self):
        self._test_copytruncate_larger(False)

    def test_copytruncate_larger_on(self):
        self._test_copytruncate_larger(True)

    def test_offset_file(self):
        pygtail = Pygtail(self.logfile.name, offset_mode='paranoid')

        log_inode = os.stat(self.logfile.name).st_ino or os.stat(self.logfile.name).st_ctime_ns

        next(pygtail)
        with open(self.logfile.name + '.offset', 'r') as f:
            inode, offset = int(next(f)), int(next(f))
        self.assertEqual(inode, log_inode)
        self.assertEqual(offset, 2)

        next(pygtail)
        with open(self.logfile.name + '.offset', 'r') as f:
            inode, offset = int(next(f)), int(next(f))
        self.assertEqual(inode, log_inode)
        self.assertEqual(offset, 4)

        next(pygtail)
        with open(self.logfile.name + '.offset', 'r') as f:
            inode, offset = int(next(f)), int(next(f))
        self.assertEqual(inode, log_inode)
        self.assertEqual(offset, 6)

    def test_manual_offset_mode(self):
        pygtail = Pygtail(self.logfile.name, offset_mode='manual')

        log_inode = os.stat(self.logfile.name).st_ino or os.stat(self.logfile.name).st_ctime_ns

        next(pygtail)
        self.assertFalse(os.path.exists(self.logfile.name + '.offset'))

        next(pygtail)
        self.assertFalse(os.path.exists(self.logfile.name + '.offset'))
        pygtail.update_offset_file()
        with open(self.logfile.name + '.offset', 'r') as f:
            inode, offset = int(next(f)), int(next(f))
        self.assertEqual(inode, log_inode)
        self.assertEqual(offset, 4)

        next(pygtail)
        del pygtail

        pygtail = Pygtail(self.logfile.name, offset_mode='manual')
        next(pygtail)
        pygtail.update_offset_file()
        with open(self.logfile.name + '.offset', 'r') as f:
            inode, offset = int(next(f)), int(next(f))
        self.assertEqual(inode, log_inode)
        self.assertEqual(offset, 6)

    def test_readlines_manual_offset(self):
        pygtail = Pygtail(self.logfile.name, offset_mode='manual')
        self.assertEqual(pygtail.readlines(), self.test_lines)

        del pygtail

        pygtail = Pygtail(self.logfile.name, offset_mode='manual')
        self.assertEqual(pygtail.readlines(), self.test_lines)


    def test_endfile_offset_mode(self):
        pygtail = Pygtail(self.logfile.name, offset_mode='endfile')

        log_inode = os.stat(self.logfile.name).st_ino or os.stat(self.logfile.name).st_ctime_ns

        pygtail.readline()
        pygtail.readline()
        pygtail.readline()
        pygtail.readline()
        self.assertFalse(os.path.exists(self.logfile.name + '.offset'))

        pygtail.readline()
        self.assertFalse(os.path.exists(self.logfile.name + '.offset'))

        pygtail.readline()
        self.assertTrue(os.path.exists(self.logfile.name + '.offset'))
        with open(self.logfile.name + '.offset', 'r') as f:
            inode, offset = int(next(f)), int(next(f))
        self.assertEqual(inode, log_inode)
        self.assertEqual(offset, 10)

    def test_on_update_with_paranoid(self):
        updates = [0]

        def record_update():
            updates[0] += 1

        pygtail = Pygtail(self.logfile.name,  offset_mode='paranoid',
                          on_update=record_update)

        self.assertEqual(updates[0], 0)
        next(pygtail)
        self.assertEqual(updates[0], 1)
        next(pygtail)
        self.assertEqual(updates[0], 2)
        next(pygtail)
        self.assertEqual(updates[0], 3)

    def test_on_update_without_paranoid(self):
        updates = [0]

        def record_update():
            updates[0] += 1

        pygtail = Pygtail(self.logfile.name, on_update=record_update)

        self.assertEqual(updates[0], 0)
        for line in pygtail:
            self.assertEqual(updates[0], 0)
        self.assertEqual(updates[0], 1)

    def test_every_n(self):
        updates = [0]
        # We save before returning the second line.
        # We save at the end of the file with all 3 recorded.
        expected = [1, 3]
        previous_lines = 0

        def record_update():
            self.assertEqual(previous_lines, expected[updates[0]])
            updates[0] += 1

        pygtail = Pygtail(self.logfile.name,  offset_mode='every', every_n=2, on_update=record_update)

        self.assertEqual(updates[0], 0)
        for line in pygtail:
            previous_lines += 1

    @unittest.skipIf(sys.platform.startswith('win'), "lock file mechanism on windows does not allow rename")
    def test_renamecreate(self):
        """
        Tests "renamecreate" semantics where the currently processed file gets renamed and the
        original file gets recreated. This is the behavior of certain logfile rollers such as
        TimeBasedRollingPolicy in Java's Logback library.
        """
        new_lines = ["4\n5\n", "6\n7\n"]
        pygtail = Pygtail(self.logfile.name)
        pygtail.read()
        os.rename(self.logfile.name, "%s.2018-03-10" % self.logfile.name)
        # append will recreate the original log file
        self.append(new_lines[0])
        self.append(new_lines[1])
        self.assertEqual(pygtail.read(), ''.join(new_lines))

    def test_full_lines(self):
        """
        Tests lines are logged only when they have a new line at the end. This is useful to ensure that log lines
        aren't unintentionally split up.
        """
        pygtail = Pygtail(self.logfile.name, full_lines=True)
        new_lines = "4\n5,"
        last_line = "5.5\n6\n"

        self.append(new_lines)
        pygtail.read()
        self.append(last_line)
        self.assertEqual(pygtail.read(), "5,5.5\n6\n")


def main():
    unittest.main(buffer=True)


if __name__ == "__main__":
    main()


import json
import typing as t
from datetime import datetime
from http.client import HTTPException

import flask
from flask import Blueprint, jsonify
from jsonschema import ValidationError
from sqlalchemy.orm.exc import NoResultFound
from werkzeug.exceptions import InternalServerError

from dm import defaults
from dm.utils.helpers import is_iterable_not_string, is_string_types
from dm.utils.typos import Id

if t.TYPE_CHECKING:
    from dm.web.network import Response
    from dm.domain.entities import Scope, State, Server

bp_errors = Blueprint('errors', __name__)


class BaseError(Exception):
    status_code = 400

    def _format_error_msg(self) -> str:
        pass

    def format(self) -> str:
        return self._format_error_msg()

    @property
    def payload(self) -> t.Optional[dict]:
        return self.__dict__

    def __str__(self):
        msg = self.format()
        payload = None
        if self.payload:
            try:
                payload = json.dumps(self.payload, indent=4)
            except:
                try:
                    payload = str(self.__dict__)
                except:
                    pass
        if msg and payload:
            return f"{msg}\n{payload}"
        else:
            return msg if msg else payload


def format_error_content(error):
    content = {
        'error': {
            'type': error.__class__.__name__,
            'message': error.format()
        }
    }

    if error.payload:
        content['error'].update(error.payload)
    return content


def format_error_response(error) -> flask.Response:
    content = format_error_content(error)
    rv = jsonify(content)
    rv.status_code = error.status_code or 500

    return rv


@bp_errors.app_errorhandler(BaseError)
def handle_error(error):
    return format_error_response(error)


@bp_errors.app_errorhandler(ValidationError)
def validation_error(error: ValidationError):
    response = {"error": {'type': error.__class__.__name__,
                          'message': error.message,
                          'path': list(error.absolute_path),
                          'schema': error.schema}}
    return response, 400


@bp_errors.app_errorhandler(HTTPException)
def handle_exception(e):
    """Return JSON instead of HTML for HTTP errors."""
    # start with the correct headers and status code from the error
    response = e.get_response()
    # replace the body with JSON
    response.data = json.dumps({'error': {'type': e.name, 'message': e.description}})
    response.content_type = "application/json"
    return response


@bp_errors.app_errorhandler(InternalServerError)
def handle_500(error):
    status_code = 500
    original = getattr(error, "original_exception", None)

    response = {'error': {
        'type': error.__class__.__name__,
        'message': error.description,
    }
    }

    # if current_app.config['DEBUG']:
    #
    #     args = [str(x) for x in error.args]
    #     if len(args) == 1:
    #         response['error']['message'] = args[0]
    #     elif len(args) > 1:
    #         response['error']['message'] = args

    return jsonify(response), status_code


class GenericError(BaseError):

    def __init__(self, message, status_code=None, **payload):
        Exception.__init__(self)
        self.message = message
        if status_code is not None:
            self.status_code = status_code
        self._payload = payload

    def _format_error_msg(self):
        return self.message

    @property
    def payload(self) -> dict:
        return self._payload


class ServerNormalizationError(BaseError):
    status_code = 404

    def __init__(self, idents: t.List[str]):
        self.idents = idents

    def _format_error_msg(self) -> str:
        return f"Servers not found in catalog: {', '.join(self.idents)}"


class EntityNotFound(BaseError):
    status_code = 404

    def __init__(self, entity: str, ident, columns: t.List[str] = None):
        self.entity = entity
        self.ident = ident
        if is_iterable_not_string(columns):
            self.columns = columns
        elif is_string_types(columns):
            self.columns = [columns]
        else:
            self.columns = ['id']

    def _format_error_msg(self) -> str:
        return f"{self.entity} with given {', '.join(self.columns)} doesn't exist"

    @property
    def payload(self) -> t.Optional[dict]:
        return dict(entity=self.entity, id=self.ident)


class NoDataFound(BaseError):
    status_code = 404

    def __init__(self, entity):
        self.entity = entity

    def _format_error_msg(self) -> str:
        return f"No data found"


class UnknownServer(BaseError):
    status_code = 404

    def __init__(self, server_id: str):
        self.server_id = server_id

    def _format_error_msg(self) -> str:
        return f"Unknown Server '{self.server_id}'"


class ObsoleteCatalog(BaseError):
    status_code = 409

    def __init__(self, actual_catalog: datetime, obsolete_catalog: datetime):
        self.actual_catalog = actual_catalog
        self.obsolete_catalog = obsolete_catalog

    def _format_error_msg(self) -> str:
        return f"Catalog is not up to date"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'current': self.actual_catalog.strftime(defaults.DATEMARK_FORMAT),
                'old': self.obsolete_catalog.strftime(defaults.DATEMARK_FORMAT)}


class CatalogMismatch(BaseError):
    status_code = 500

    def __init__(self, local_entities, remote_entities):
        self.local_entities = local_entities
        self.remote_entities = remote_entities

    def _format_error_msg(self) -> str:
        return "List entities do not match"

#################
# Locker Errors #
#################

class LockerError(BaseError):
    status_code = 409
    action_map = {'L': 'lock', 'U': 'unlock', 'P': 'prevent'}

    def __init__(self, scope):
        self.scope = scope

    @property
    def payload(self) -> t.Optional[dict]:
        return {'scope': self.scope.name}


class PriorityLocker(LockerError):
    def _format_error_msg(self) -> str:
        return 'Priority locker acquired'


class ApplicantLockerError(LockerError):
    def _format_error_msg(self) -> str:
        return f'Another applicant has the scope acquired'


class StatusLockerError(LockerError):

    def __init__(self, scope: 'Scope', action: str, current_state: 'State'):
        super().__init__(scope)
        self.action = self.action_map.get(action, action)
        self.current_state = current_state

    def _format_error_msg(self) -> str:
        return f"Unable to perform locker state transition"

    @property
    def payload(self) -> t.Optional[dict]:
        p = super().payload
        p.update(action=self.action, state=self.current_state.name)
        return p


class LockError(LockerError):

    def __init__(self, scope: 'Scope', action: str, responses: t.List['Response']):
        self.scope = scope
        self.action = self.action_map.get(action, action)
        self.responses = responses

    def _format_error_msg(self) -> str:
        return f"Unable to {self.action} on scope {self.scope.name}"

    @property
    def payload(self) -> t.Optional[dict]:
        responses = []
        for r in self.responses:
            data = dict({'id': str(r.server.id), 'name': r.server.name})
            if r.code:
                data.update(response=r.msg, code=r.code)
            else:
                if isinstance(r.exception, BaseError):
                    data.update(format_error_content(r.exception))
                else:
                    data.update(error=str(r.exception) if str(r.exception) else r.exception.__class__.__name__)
            responses.append(data)
        return {'servers': responses}

    @property
    def status_code(self):
        codes = set([r.code for r in self.responses])
        if len(codes) == 1:
            return codes.pop()
        else:
            if any(map(lambda c: c is None or c >= 500, list(codes))):
                return 500
            else:
                return super().status_code


########################
# Orchestration Errors #
########################
class TargetUnspecified(BaseError):
    status_code = 404

    def __init__(self, target: t.Iterable[str]):
        self.target = list(target)

    def _format_error_msg(self) -> str:
        return f"Target not specified"


class TargetNotNeeded(BaseError):
    status_code = 400

    def __init__(self, target: t.Iterable[str]):
        self.target = list(target)

    def _format_error_msg(self) -> str:
        return f"Target not in orchestration"


class EmptyTarget(BaseError):
    status_code = 400

    def __init__(self, target: str):
        self.target = target

    def _format_error_msg(self) -> str:
        return f"Target does not have any host specified"


class DuplicatedId(BaseError):
    status_code = 400

    def __init__(self, rid):
        self.rid = rid

    def _format_error_msg(self) -> str:
        return "Id already exists"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'id': self.rid}


class UndoStepWithoutParent(BaseError):

    def __init__(self, step_id: Id) -> None:
        self.step_id = step_id

    def _format_error_msg(self) -> str:
        return "'undo' step must have a parent step"


class ParentUndoError(BaseError):

    def __init__(self, step_id: Id, parent_step_ids: t.Union[Id, t.List[Id]]) -> None:
        if isinstance(parent_step_ids, str):
            self.parent_step_ids = [parent_step_ids]
        else:
            self.parent_step_ids = parent_step_ids
        self.step_id = step_id

    def _format_error_msg(self) -> str:
        return "a 'do' step cannot have parent 'undo' steps"


class ChildDoError(BaseError):

    def __init__(self, step_id: Id, child_step_ids: t.Union[Id, t.List[Id]]) -> None:
        if isinstance(child_step_ids, str):
            self.child_step_ids = [child_step_ids]
        else:
            self.child_step_ids = child_step_ids
        self.step_id = step_id

    def _format_error_msg(self) -> str:
        return "an 'undo' step cannot have child 'do' steps"


class CycleError(BaseError):

    def _format_error_msg(self) -> str:
        return "Cycle detected while trying to add dependency"

##############
# Deployment #
##############

class Timeout(BaseError):
    status_code = 500

class RemoteServerTimeout(BaseError):
    status_code = 500

    def __init__(self, timeout, server, command):
        self.timeout = timeout
        self.server = server
        self.command = command

    def _format_error_msg(self) -> str:
        return "Timeout reached waiting remote server operation completion"


#############
# Send File #
#############

class SoftwareServerNotFound(BaseError):
    status_code = 404

    def __init__(self, software_id, server_id):
        self.software_id = software_id
        self.server_id = server_id

    def _format_error_msg(self) -> str:
        return "Software Server Association not found"


class NoSoftwareServer(BaseError):
    status_code = 500

    def __init__(self, software_id):
        self.software_id = software_id

    def _format_error_msg(self) -> str:
        return "Software does not have any server location"


class ChunkSendError(BaseError):
    status_code = 500

    def __init__(self, chunk_responses: t.Dict[int, 'Response']):
        self.chunk_responses = chunk_responses

    def _format_error_msg(self) -> str:
        return "Error while trying to send chunks to server"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'chunks': {c: r.to_dict() for c, r in self.chunk_responses.items()}}


class TransferBase(BaseError):
    status_code = 409


class TransferFileAlreadyExists(TransferBase):

    def __init__(self, file=None):
        self.file = file

    def _format_error_msg(self) -> str:
        return "File already exists. Use force=True if needed"


class TransferFileAlreadyOpen(TransferBase):

    def __init__(self, file=None):
        self.file = file

    def _format_error_msg(self) -> str:
        return "There is already a transfer sending file"


class TransferSoftwareAlreadyOpen(TransferBase):

    def __init__(self, software_id=None):
        self.software_id = software_id

    def _format_error_msg(self) -> str:
        return "There is already a transfer sending software"


class TransferNotInValidState(TransferBase):
    status_code = 410

    def __init__(self, transfer_id: str, status: str):
        self.status = status
        self.transfer_id = transfer_id

    def _format_error_msg(self) -> str:
        return "Transfer not in a valid state"


#################
# Common Errors #
#################
class UnreachableDestination(BaseError):
    status_code = 503

    def __init__(self, server: 'Server', proxy: 'Server' = None):
        from dm.domain.entities import Server
        try:
            self.proxy = proxy or Server.get_current()
        except NoResultFound:
            self.proxy = None

        self.server = server

    def _format_error_msg(self) -> str:
        return f"Unreachable destination"

    @property
    def payload(self) -> t.Optional[dict]:
        data = dict(destination=dict(name=self.server.name, id=str(self.server.id)))
        if self.proxy:
            data.update(proxy=dict(name=self.proxy.name, id=str(self.proxy.id)))
        return data


class KeywordReserved(BaseError):
    status_code = 400

    def __init__(self, msg):
        self.msg = msg

    def _format_error_msg(self) -> str:
        return self.msg


class FileNotFound(BaseError):
    status_code = 404

    def __init__(self, file):
        self.file = file

    def _format_error_msg(self) -> str:
        return "File not found"


class HTTPError(BaseError):

    def __init__(self, resp: 'Response'):
        self.resp = resp
        self.status_code = resp.code if resp.code else 500

    def _format_error_msg(self) -> str:
        return "Error on request"

    @property
    def payload(self) -> t.Optional[dict]:
        return self.resp.to_dict()


class ProxyForwardingError(BaseError):
    status_code = 502

    def __init__(self, dest, exception):
        self.dest = dest
        self.exception = exception

    def _format_error_msg(self) -> str:
        return "Error while trying to forward request"

    @property
    def payload(self) -> t.Optional[dict]:
        return {'server': {'id': str(self.dest.id), 'name': self.dest.name},
                'exception': str(self.exception) if str(self.exception) else self.exception.__class__.__name__}


class HealthCheckMismatch(BaseError):
    status_code = 500

    def __init__(self, expected: t.Dict[str, str], actual: t.Dict[str, str]):
        self.expected = expected
        self.actual = actual

    def _format_error_msg(self) -> str:
        return "Healtcheck response does not match with the server requested"


class ParameterMustBeSet(BaseError):
    status_code = 404

    def __init__(self, msg):
        self.msg = msg

    def _format_error_msg(self) -> str:
        return self.msg



import typing as t

from dm.network.low_level import check_host
from dm.utils.typos import UUID
from dm.web import db

if t.TYPE_CHECKING:
    from dm.domain.entities import Server, Gate


class Route(db.Model):
    __tablename__ = 'L_route'

    destination_id = db.Column(UUID, db.ForeignKey('D_server.id'), primary_key=True, nullable=False)
    proxy_server_id = db.Column(UUID, db.ForeignKey('D_server.id'))
    gate_id = db.Column(UUID, db.ForeignKey('D_gate.id'))
    cost = db.Column(db.Integer)

    destination = db.relationship("Server", foreign_keys=[destination_id], back_populates="route")
    proxy_server = db.relationship("Server", foreign_keys=[proxy_server_id])
    gate = db.relationship("Gate", foreign_keys=[gate_id])

    def __init__(self, destination: 'Server', proxy_server: 'Server' = None, gate: 'Gate' = None, cost: int = None):
        self.destination = destination
        if proxy_server:
            if proxy_server == destination:
                raise ValueError('You must specify a gate when proxy_server equals destination')
            else:
                if cost is None or cost == 0:
                    raise ValueError("Cost must be specified and greater than 0 when proxy_server")
                self.proxy_server = proxy_server
                self.cost = cost
        elif gate:
            # check if gate is from neighbour or from a proxy server
            if gate and destination == gate.server:
                if cost is not None and cost > 0:
                    raise ValueError("Cost must be set to 0 when defining route for a neighbour")
                self.gate = gate
                self.cost = 0
            else:
                if cost is None or cost <= 0:
                    raise ValueError("Cost must be specified and greater than 0 when gate from a proxy_server")
                else:
                    self.gate = gate
                    self.cost = cost
        elif cost == 0:
            # find a gateway and set that gateway as default
            if len(destination.external_gates) == 1:
                self.gate = destination.external_gates[0]
                self.cost = 0
            else:
                for gate in destination.external_gates:
                    if check_host(gate.dns or str(gate.ip), gate.port, timeout=1, retry=3, delay=0.5):
                        self.gate = gate
                        self.cost = 0
                        break
        # if not (self.gate or self.proxy_server):
        #     raise ValueError('Not a valid route')

    def __str__(self):
        gate = f"{self.gate.server}://{self.gate}" if self.gate else "None"
        return f"{self.destination} -> " \
               f"proxy_server={self.proxy_server}, " \
               f"gate={gate}, " \
               f"cost={self.cost}"

    def __repr__(self):
        return f"Route({self.to_json()})"

    def to_json(self, human=False):
        if not self.destination.id or (self.proxy_server and not self.proxy_server.id) or (
                self.gate and not self.gate.id):
            raise RuntimeError("commit object before dump to json")
        if human:
            return {'destination': str(self.destination) if self.destination else None,
                    'proxy_server': str(self.proxy_server) if self.proxy_server else None,
                    'gate': str(self.gate) if self.gate else None,
                    'cost': self.cost}
        else:
            return {'destination_id': str(getattr(self.destination, 'id', '')) or None,
                'proxy_server_id': str(getattr(self.proxy_server, 'id', '')) or None,
                'gate_id': str(getattr(self.gate, 'id', '')) or None,
                'cost': self.cost}



import uuid
from datetime import datetime

from dm.domain.entities.base import DistributedEntityMixin, EntityReprMixin
from dm.utils.helpers import get_now
from dm.utils.typos import UUID, Kwargs, UtcDateTime
from dm.web import db


class ServiceOrchestration(db.Model, EntityReprMixin, DistributedEntityMixin):
    __tablename__ = 'D_service_orchestration'
    order = 50
    id = db.Column(UUID, primary_key=True)
    service_id = db.Column(UUID, db.ForeignKey('D_service.id'))
    orchestration_id = db.Column(UUID, db.ForeignKey('D_orchestration.id'))
    execution_time = db.Column(UtcDateTime(), default=get_now())


class Service(db.Model, EntityReprMixin, DistributedEntityMixin):
    __tablename__ = 'D_service'
    order = 40

    id = db.Column(UUID, primary_key=True, default=uuid.uuid4)
    name = db.Column(db.String(255), nullable=False)
    details = db.Column(db.JSON)
    created_on = db.Column(UtcDateTime(timezone=True), nullable=False, default=get_now)
    last_ping = db.Column(UtcDateTime(timezone=True))
    status = db.Column(db.String(40))

    orchestrations = db.relationship("Orchestration", secondary="D_service_orchestration", order_by="ServiceOrchestration.execution_time")

    def __init__(self, name: str, details: Kwargs, status: str, created_on: datetime = get_now(),
                 last_ping: datetime = None, id: uuid.UUID = None, **kwargs):
        DistributedEntityMixin.__init__(self, **kwargs)
        self.id = id
        self.name = name
        self.details = details
        self.status = status
        self.created_on = created_on
        self.last_ping = last_ping

    def to_json(self):
        return {'id': str(self.id), 'name': self.name, 'details': self.details,
                'last_ping': self.last_ping.strftime("%d/%m/%Y %H:%M:%S"),
                'status': self.status}



from dataclasses import dataclass


@dataclass
class Token:
    """
    Class that saves relation between a sent message and its asynchronous return.
    """
    id: int
    source: str
    destination: str

    @property
    def uid(self):
        return self.source + '.' + self.destination + '.' + str(self.id)






import base64
from unittest import TestCase, mock
from unittest.mock import patch

from aioresponses import aioresponses, CallbackResult
from asynctest.mock import patch as async_patch

from dm.domain.entities import Server, Route, Log, User
from dm.use_cases.log_sender import LogSender
from dm.use_cases.log_sender import _PygtailBuffer, Pygtail
from dm.utils.asyncio import run
from dm.web import create_app, db


class TestServer(TestCase):

    def setUp(self):
        """Create and configure a new app instance for each test."""
        # create the app with common test config
        self.app = create_app('test')
        self.app_context = self.app.app_context()
        self.app_context.push()
        db.create_all()
        User.set_initial()
        self.source = Server('source', port=8000, me=True)
        self.dest = Server('dest', port=8000)
        Route(self.dest, cost=0)
        db.session.add_all([self.source, self.dest])
        _PygtailBuffer._fh = mock.MagicMock()
        self.log_sender = LogSender()

    def tearDown(self) -> None:
        db.session.remove()
        db.drop_all()
        self.app_context.pop()

    @patch('dm.use_cases.log_sender.os.walk', autospec=True)
    @patch('dm.use_cases.log_sender.os.path.isfile', autospec=True)
    @patch.object(Pygtail, 'update_offset_file')
    @patch.object(_PygtailBuffer, 'readlines', return_value='content', autospec=True)
    @aioresponses()
    def test_log_sender_file(self, mock_pb_rl, mock_pt_uof, mock_isfile, mock_walk, m):
        def callback(url, **kwargs):
            self.assertDictEqual(
                {"file": '/dm/logs/dm.log', 'data': base64.b64encode('content'.encode()).decode('ascii')},
                kwargs['json'])
            return CallbackResult('POST', status=200)

        m.post(self.dest.url('api_1_0.logresource', log_id='aaaaaaaa-1234-5678-1234-56781234aaa1'), callback=callback)

        mock_isfile.return_value = True

        log = Log(id='aaaaaaaa-1234-5678-1234-56781234aaa1', source_server=self.source, target='/var/log/dm.log',
                  destination_server=self.dest, dest_folder='/dm/logs/')
        db.session.add(log)

        run(self.log_sender.send_new_data())

        mock_isfile.assert_called_once()
        mock_walk.assert_not_called()
        mock_pb_rl.assert_called_once()
        mock_pt_uof.assert_called_once()

    @patch('dm.use_cases.log_sender.os.path.isfile', autospec=True)
    @patch.object(Pygtail, 'update_offset_file')
    @patch.object(_PygtailBuffer, 'readlines', return_value='content', autospec=True)
    @aioresponses()
    def test_log_sender_file_no_dest_folder(self, mock_pb_rl, mock_pt_uof, mock_isfile, m):
        def callback(url, **kwargs):
            self.assertDictEqual(
                {"file": '/var/log/dm.log', 'data': base64.b64encode('content'.encode()).decode('ascii')},
                kwargs['json'])
            return CallbackResult('POST', status=200)

        m.post(self.dest.url('api_1_0.logresource', log_id='aaaaaaaa-1234-5678-1234-56781234aaa1'), callback=callback)

        mock_isfile.return_value = True

        log = Log(id='aaaaaaaa-1234-5678-1234-56781234aaa1', source_server=self.source, target='/var/log/dm.log',
                  destination_server=self.dest, dest_folder=None)
        db.session.add(log)

        run(self.log_sender.send_new_data())

        mock_isfile.assert_called_once()
        mock_pb_rl.assert_called_once()
        mock_pt_uof.assert_called_once()

    @async_patch('dm.use_cases.log_sender.async_post', autospec=True)
    @patch('dm.use_cases.log_sender.os.walk', autospec=True)
    @patch('dm.use_cases.log_sender.os.path.isfile', autospec=True)
    @patch.object(Pygtail, 'update_offset_file')
    @patch.object(_PygtailBuffer, 'readlines', side_effect=['content1', 'newcontent2'], autospec=True)
    @aioresponses()
    def test_log_sender_folder(self, mock_pb_rl, mock_pt_uof, mock_isfile, mock_walk, mock_post, m):

        def callback(url, **kwargs):
            if kwargs['json']['file'] == '/dm/logs/log1':
                self.assertDictEqual(
                    {"file": '/dm/logs/log1', 'data': base64.b64encode('content1'.encode()).decode('ascii')},
                    kwargs['json'])
                return CallbackResult('POST', payload={'offset': 8}, status=200)
            elif kwargs['json']['file'] == '/dm/logs/dir1/log2':
                self.assertDictEqual(
                    {"file": '/dm/logs/dir1/log2', 'data': base64.b64encode('newcontent2'.encode()).decode('ascii')},
                    kwargs['json'])
                return CallbackResult('POST', payload={'offset': 11}, status=200)
            else:
                raise

        mock_post.side_effect = [({'offset': 8}, 200), ({'offset': 11}, 200), ({'offset': 8}, 200)]

        m.post(self.dest.url('api_1_0.logresource', log_id='aaaaaaaa-1234-5678-1234-56781234aaa1'), callback=callback)
        m.post(self.dest.url('api_1_0.logresource', log_id='aaaaaaaa-1234-5678-1234-56781234aaa1'), callback=callback)
        m.post(self.dest.url('api_1_0.logresource', log_id='aaaaaaaa-1234-5678-1234-56781234aaa1'), callback=callback)

        mock_isfile.return_value = False
        mock_walk.side_effect = [
            [('/var/log', ['dir1'], ['log1', 'file']), ('/var/log/dir1', ['dir2'], ['log2'])],
            [('/var/log', ['dir1'], ['log1', 'file']), ('/var/log/dir1', ['dir2'], [])]
        ]

        log = Log(id='aaaaaaaa-1234-5678-1234-56781234aaa1', source_server=self.source, target='/var/log',
                  destination_server=self.dest, dest_folder='/dm/logs/', include='^(log|dir)', exclude='^dir2',
                  recursive=True)
        db.session.add(log)

        run(self.log_sender.send_new_data())

        mock_isfile.assert_called_once()
        self.assertEqual(2, mock_pb_rl.call_count)
        self.assertEqual(2, mock_pt_uof.call_count)



from flask import url_for

from dm.domain.entities import Server, Gate
from dm.domain.entities.route import Route
from dm.web import db
from tests.helpers import TestDimensigonBase


class TestRoutes(TestDimensigonBase):


    def test_routes_get(self):
        s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='server1', me=True)
        g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
                  dns=s1.name)
        s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='server2')
        g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
                  dns=s2.name)
        Route(s2, gate=g2, cost=0)
        s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='server3')
        g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
                  dns=s3.name)
        Route(s3, gate=g3, cost=0)
        s4 = Server(id='123e4567-e89b-12d3-a456-426655440004', name='server4')
        g4 = Gate(id='123e4567-e89b-12d3-a456-426655440014', server=s4, port=5001,
                  dns=s4.name)
        Route(s4, proxy_server=s2, cost=1)
        db.session.add_all([s1, s2, s3, s4])
        db.session.commit()

        response = self.client.get(url_for('api_1_0.routes', _external=False),
                                   headers=self.auth.header)
        data = response.get_json()
        self.assertDictEqual({'server_id': '123e4567-e89b-12d3-a456-426655440001',
                              'route_list': [
                                  dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
                                       gate_id='123e4567-e89b-12d3-a456-426655440012',
                                       proxy_server_id=None, cost=0),
                                  dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
                                       gate_id='123e4567-e89b-12d3-a456-426655440013',
                                       proxy_server_id=None, cost=0),
                                  dict(destination_id='123e4567-e89b-12d3-a456-426655440004',
                                       gate_id=None,
                                       proxy_server_id='123e4567-e89b-12d3-a456-426655440002', cost=1)]}, data)

    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch(self, mocked_ping, mocked_thread):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='server1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='server2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     Route(s2, gate=g2, cost=0)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='server3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #     Route(s3, gate=g3, cost=0)
    #     s4 = Server(id='123e4567-e89b-12d3-a456-426655440004', name='server4')
    #     g4 = Gate(id='123e4567-e89b-12d3-a456-426655440014', server=s4, port=5001,
    #               dns=s4.name)
    #     Route(s4, proxy_server=s2, cost=1)
    #     db.session.add_all([s1, s2, s3, s4])
    #
    #     mocked_ping.return_value = (None, None)
    #     access_token = create_access_token(identity='test')
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers={"Authorization": f"Bearer {access_token}"},
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440004',
    #                                                 gate_id=None,
    #                                                 proxy_server_id=None,
    #                                                 cost=None)]})
    #
    #     # s = Server.query.get('123e4567-e89b-12d3-a456-426655440001')
    #     # self.assertEqual(None, s.route.gate)
    #     # self.assertEqual(None, s.route.proxy_server)
    #     # self.assertEqual(None, s.route.cost)
    #
    #     # s = Server.query.get('123e4567-e89b-12d3-a456-426655440002')
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     # s = Server.query.get('123e4567-e89b-12d3-a456-426655440003')
    #     self.assertEqual(g3, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(0, s3.route.cost)
    #
    #     # s = Server.query.get('123e4567-e89b-12d3-a456-426655440004')
    #     self.assertEqual(None, s4.route.gate)
    #     self.assertEqual(None, s4.route.proxy_server)
    #     self.assertEqual(None, s4.route.cost)
    #
    #     self.assertEqual(1, mocked_thread.Thread.call_count)
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.update_table_routing_cost')
    # @responses.activate
    # def test_routes_post(self, mocked_utr, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='server1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='server2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5001,
    #               dns=s2.name)
    #     Route(destination=s2, cost=0)
    #     db.session.add_all([s1, s2])
    #
    #     mocked_utr.return_value = {s1: TempRoute(proxy_server=s1, gate=g1, cost=0)}
    #
    #     resp = self.client.post(url_for('api_1_0.routes'),
    #                             json={'discover_new_neighbours': True, 'check_current_neighbours': True},
    #                             headers=self.headers)
    #
    #     mocked_utr.assert_called_once_with(discover_new_neighbours=True, check_current_neighbours=True)
    #
    #     args, kwargs = mocked_threading.Thread.call_args
    #
    #     self.assertTupleEqual((self.app, s2, 'api_1_0.routes'), kwargs['args'])
    #
    #     self.assertDictEqual({'server_id': '123e4567-e89b-12d3-a456-426655440001',
    #                           'route_list': [{'destination_id': '123e4567-e89b-12d3-a456-426655440001',
    #                                           'proxy_server_id': '123e4567-e89b-12d3-a456-426655440001',
    #                                           'gate_id': '123e4567-e89b-12d3-a456-426655440011',
    #                                           'cost': 0}]}, kwargs['kwargs']['json'])
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario1(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     Route(s2, gate=g2, cost=0)
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return 0, None
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return True
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440003',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440011',
    #                                                 proxy_server_id=None,
    #                                                 cost=0),
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440012',
    #                                                 proxy_server_id=None,
    #                                                 cost=0)
    #                                        ]})
    #
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(g3, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(0, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_called_once()
    #     args, kwargs = mocked_threading.Thread.call_args
    #     kwargs.pop('target')
    #     self.assertDictEqual(dict(args=(self.app, s2, 'api_1_0.routes'),
    #                               kwargs={'json': {'server_id': str(s1.id),
    #                                                'route_list': [
    #                                                    {'destination_id': str(s3.id),
    #                                                     'proxy_server_id': None,
    #                                                     'gate_id': str(g3.id),
    #                                                     'cost': 0}]},
    #                                       'headers': self.headers}), kwargs)
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario2(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     Route(s2, gate=g2, cost=0)
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return None, None
    #         else:
    #             raise
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return False
    #         else:
    #             raise
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440013',
    #                                                 proxy_server_id=None,
    #                                                 cost=0),
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440002',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440012',
    #                                                 proxy_server_id=None,
    #                                                 cost=0)
    #                                        ]})
    #
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(None, s3.route.gate)
    #     self.assertEqual(s2, s3.route.proxy_server)
    #     self.assertEqual(1, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_not_called()
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario3(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     Route(s2, gate=g2, cost=0)
    #     Route(s3, proxy_server=s2, cost=1)
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return 1, None
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return True
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440003',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440011',
    #                                                 proxy_server_id=None,
    #                                                 cost=0)
    #                                        ]})
    #
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(g3, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(0, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_called_once()
    #     args, kwargs = mocked_threading.Thread.call_args
    #     kwargs.pop('target')
    #     self.assertDictEqual(dict(args=(self.app, s2, 'api_1_0.routes'),
    #                               kwargs={'json': {'server_id': str(s1.id),
    #                                                'route_list': [
    #                                                    {'destination_id': str(s3.id),
    #                                                     'proxy_server_id': None,
    #                                                     'gate_id': str(g3.id),
    #                                                     'cost': 0}]},
    #                                       'headers': self.headers}), kwargs)
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario4(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return None, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return None, None
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return False
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440001',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440011',
    #                                                 proxy_server_id=None,
    #                                                 cost=0),
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
    #                                                 gate_id='123e4567-e89b-12d3-a456-426655440013',
    #                                                 proxy_server_id=None,
    #                                                 cost=0)
    #                                        ]})
    #
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(None, s3.route.gate)
    #     self.assertEqual(s2, s3.route.proxy_server)
    #     self.assertEqual(1, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_not_called()
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario5(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     Route(s2, gate=g2, cost=0)
    #     Route(s3, gate=g3, cost=0)
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return 0, None
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return True
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
    #                                                 gate_id=None,
    #                                                 proxy_server_id=None,
    #                                                 cost=None)
    #                                        ]})
    #
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(g3, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(0, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_not_called()
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario6(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     Route(s2, gate=g2, cost=0)
    #     Route(s3, gate=g3, cost=0)
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return None, None
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return False
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
    #                                                 gate_id=None,
    #                                                 proxy_server_id=None,
    #                                                 cost=None)
    #                                        ]})
    #
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(None, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(None, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_not_called()
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario7(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5001,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5002,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5003,
    #               dns=s3.name)
    #
    #     Route(s2, proxy_server=None, gate=g2, cost=0)
    #     Route(s3, proxy_server=s2, gate=None, cost=1)
    #     db.session.add_all([s1, s2, s3])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return None, None
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return False
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440003',
    #                                                 gate_id=None,
    #                                                 proxy_server_id=None,
    #                                                 cost=None)
    #                                        ]})
    #
    #     self.assertEqual(0, mocked_ping.call_count)
    #     self.assertEqual(0, mocked_check_host.call_count)
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(None, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(None, s3.route.cost)
    #
    #     mocked_threading.Thread.assert_not_called()
    #
    # @patch('dm.web.api_1_0.urls.use_cases.threading')
    # @patch('dm.web.api_1_0.urls.use_cases.check_host')
    # @patch('dm.web.api_1_0.urls.use_cases.ping_server')
    # def test_routes_patch_scenario8(self, mocked_ping, mocked_check_host, mocked_threading):
    #     s1 = Server(id='123e4567-e89b-12d3-a456-426655440001', name='node1', me=True)
    #     g1 = Gate(id='123e4567-e89b-12d3-a456-426655440011', server=s1, port=5000,
    #               dns=s1.name)
    #     s2 = Server(id='123e4567-e89b-12d3-a456-426655440002', name='node2')
    #     g2 = Gate(id='123e4567-e89b-12d3-a456-426655440012', server=s2, port=5000,
    #               dns=s2.name)
    #     s3 = Server(id='123e4567-e89b-12d3-a456-426655440003', name='node3')
    #     g3 = Gate(id='123e4567-e89b-12d3-a456-426655440013', server=s3, port=5000,
    #               dns=s3.name)
    #
    #     s4 = Server(id='123e4567-e89b-12d3-a456-426655440004', name='node4')
    #     g4 = Gate(id='123e4567-e89b-12d3-a456-426655440014', server=s4, port=5000,
    #               dns=s4.name)
    #
    #     Route(s2, proxy_server=None, gate=g2, cost=0)
    #     Route(s3, proxy_server=s2, gate=None, cost=1)
    #     Route(s4, proxy_server=s2, gate=None, cost=1)
    #     db.session.add_all([s1, s2, s3, s4])
    #     db.session.commit()
    #
    #     def ping(server, *args, **kwargs):
    #         if str(server.id) == '123e4567-e89b-12d3-a456-426655440001':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440002':
    #             return 0, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440003':
    #             return 1, None
    #         elif str(server.id) == '123e4567-e89b-12d3-a456-426655440004':
    #             return None, None
    #         else:
    #             raise
    #
    #     mocked_ping.side_effect = ping
    #
    #     def check_host(host, port, *args, **kwargs):
    #         if host == 'node1':
    #             return True
    #         elif host == 'node2':
    #             return True
    #         elif host == 'node3':
    #             return False
    #         elif host == 'node4':
    #             return False
    #         else:
    #             raise
    #
    #     mocked_check_host.side_effect = check_host
    #
    #     response = self.client.patch(url_for('api_1_0.routes'),
    #                                  headers=self.headers,
    #                                  json={"server_id": '123e4567-e89b-12d3-a456-426655440002',
    #                                        "route_list": [
    #                                            dict(destination_id='123e4567-e89b-12d3-a456-426655440004',
    #                                                 gate_id=None,
    #                                                 proxy_server_id=s3,
    #                                                 cost=1),
    #
    #                                        ]})
    #
    #     self.assertEqual(0, mocked_ping.call_count)
    #     self.assertEqual(0, mocked_check_host.call_count)
    #     self.assertIsNone(s1.route)
    #
    #     self.assertEqual(g2, s2.route.gate)
    #     self.assertEqual(None, s2.route.proxy_server)
    #     self.assertEqual(0, s2.route.cost)
    #
    #     self.assertEqual(g3, s3.route.gate)
    #     self.assertEqual(None, s3.route.proxy_server)
    #     self.assertEqual(0, s3.route.cost)
    #
    #     self.assertEqual(None, s4.route.gate)
    #     self.assertEqual(s3, s4.route.proxy_server)
    #     self.assertEqual(1, s4.route.cost)
    #
    #     mocked_threading.Thread.assert_called_once()
    #     args, kwargs = mocked_threading.Thread.call_args
    #     kwargs.pop('target')
    #     self.assertDictEqual(dict(args=(self.app, s2, 'api_1_0.routes'),
    #                               kwargs={'json': {'server_id': str(s1.id),
    #                                                'route_list': [
    #                                                    {'destination_id': str(s3.id),
    #                                                     'proxy_server_id': None,
    #                                                     'gate_id': str(g3.id),
    #                                                     'cost': 0},
    #                                                    {'destination_id': str(s4.id),
    #                                                     'proxy_server_id': str(s3.id),
    #                                                     'gate_id': None,
    #                                                     'cost': 1}
    #                                                ]},
    #                                       'headers': self.headers}), kwargs)


